#ifndef BOARD_H
#define BOARD_H

#include "assert.h"
#include "bits64.h"
#include "pieces.h"
#include "move.h"

#define RANKS     12
#define FILES     12
#define MAX_LEAP   3
#if 1
/* This representation tests as a little faster, probably because the number of
 * files is now a power of 2 (16). This removes the need for padding at the end of the table.
 * The last valid square index is 254, so valid squares still fit inside an unsigned char.
 *
 * NOTE: the mailbox allows for (3,2) leapers, but (3,3) leapers will still
 * leap outside the board, so those should not be added!
 */
#   define MAILBOX_RANKS (RANKS + 2*MAX_LEAP)
#   define MAILBOX_FILES (FILES + MAX_LEAP+1)
#   define MAILBOX_SIZE  (MAILBOX_RANKS * MAILBOX_FILES)
#else
#   define MAILBOX_RANKS (RANKS + 2*MAX_LEAP)
#   define MAILBOX_FILES (FILES + MAX_LEAP)
#   define MAILBOX_SIZE  (MAILBOX_RANKS * MAILBOX_FILES + MAX_LEAP)
#endif
#define MAILBOX_RANK0 MAX_LEAP
#define MAILBOX_FILE0 ((MAX_LEAP+1)/2)
#define MAILBOX_FIRST (MAX_LEAP*MAILBOX_FILES + MAILBOX_FILE0)
#define B 0xff

typedef struct piecelist_entry_t {
   uint8_t id;
   uint8_t where;
   bool virgin;
} piecelist_entry_t;

typedef struct board_t {
   piece_t mailbox[MAILBOX_SIZE];
   side_t side_to_move;
   int fifty_counter;

   piecelist_entry_t piece_list[2*MAX_PIECES_PER_SIDE];
   uint64_t live_pieces[NUM_SIDES];
   uint64_t hash;
   bool check;
} board_t;

typedef struct {
   uint64_t live[NUM_SIDES];
   uint64_t hash;
   int fifty_counter;
   piece_t piece, victim;
   piece_t piece_id, victim_id;
   bool virgin;
   bool check;
} unmake_info_t;

extern int square_to_mailbox[FILES*RANKS];
extern int mailbox_to_file[MAILBOX_SIZE];
extern int mailbox_to_rank[MAILBOX_SIZE];
extern char promotion_map[MAILBOX_FILES];
extern char square_str[MAILBOX_SIZE][4];
extern int relative_square[NUM_SIDES][MAILBOX_SIZE];

static inline int relative_rank(const board_t *board, int rank, side_t side)
{
   return (side == WHITE) ? rank : RANKS-1-rank;
}

void clear_board(board_t *board);
int generate_piece_list_index(uint64_t mask, piece_t id);
int add_piece_to_list(board_t *board, side_t side, piece_t piece, int square);
void setup_promotions(const board_t *board);

side_t piece_colour_on_square(const board_t *board, int square);
int piece_id_on_square(const board_t *board, int square);

bool is_capture_move(const board_t *board, move_t move);
bool is_promotion_move(const board_t *board, move_t move);
bool is_irreversible_move(const board_t *board, move_t move);
int get_moved_piece(const board_t *board, move_t move);
int get_move_piece(const board_t *board, move_t move);

void makemove(board_t *board, move_t move, unmake_info_t *ui);
void unmakemove(board_t *board, move_t move, unmake_info_t *ui);

int rank_file_to_mailbox(int r, int f);
side_t get_side_to_move(const board_t *board);
void set_side_to_move(board_t *board, side_t side);

#endif

