#ifndef GAME_H
#define GAME_H

#include "board.h"
#include "timer.h"
#include "move.h"
#include "movelist.h"
#include "hashtable.h"
#include "evalhash.h"

#define MAX_SEARCH_DEPTH 64       /* maximum depth of search tree */

#define HARD_HORIZON -20

#define MAX_TOTAL_DEPTH (MAX_SEARCH_DEPTH - HARD_HORIZON)

#define HASH_TABLE_SIZE (1024*1024)

typedef struct game_t {
   board_t board;      /* pointer to the current board position */

   char *start_fen;     /* FEN string encoding the starting position of this game */
   char *xb_setup;      /* setup string that has to be sent to XBoard (startup position will be attached) */

   size_t moves_played; /* Number of moves played to current position */
   size_t last_move;    /* Number of the last move played in the game; useful when we take back a move */
   int start_move_count;

   size_t max_moves;    /* Maximum number of moves that can be stored */
   int16_t *score;      /* Evaluation score associated with each position */
   move_t *move_list;   /* list of moves played in the game */
   int *move_clock;

   /* State information */
   unmake_info_t *ui;
   bool *in_check;

   movelist_t movelist[MAX_TOTAL_DEPTH];

   /* Killer moves, storage space requirements must come from the search
    * function.
    */
   move_t killer[2][MAX_TOTAL_DEPTH];
   move_t mate_killer[MAX_TOTAL_DEPTH];

   /* Botwinnik-Markov extension.
    * This is like a killer move or a refutation move, but for NULL moves.
    * When encountering it on consecutive plies, extend the search (or at least sort the move higher in the
    * list).
    */
   move_t null_killer[MAX_TOTAL_DEPTH];

   /* Counter moves, for use with the counter move heuristic.
    * Indexed by from- and to square and side to move.
    */
   move_t counter[256][256][NUM_SIDES];

   int history[NUM_SIDES][MAX_PIECE_TYPES][256];
   int max_history;


   /* Data structure for retrieving the principle variation. At each depth,
    * there are two branches for the tree: the principle variation and the
    * "current" branch. If the current branch turns out to be better than
    * the PV, it becomes the new PV.
    * A binary tree would be enough to store this information (and be easy
    * to manipulate), but for now we use a plain NxN array (of which half
    * the space is wasted, obviously)
    */
   move_t principle_variation[MAX_TOTAL_DEPTH][MAX_TOTAL_DEPTH];
   int length_of_variation[MAX_TOTAL_DEPTH];
   move_t best_move[MAX_TOTAL_DEPTH];

   /* Transposition table */
   size_t default_hash_size;
   hash_table_t *transposition_table;

   /* Evaluation hash table */
   eval_hash_table_t *eval_hash;

   /* Hash table for repetition detection */
   int8_t repetition_hash_table[0x10000];

   eval_t resign_threshold;   /* Increment resign count if the evaluation is below this limit */
   eval_t draw_threshold;     /* Increment draw count if the absolute evaluation is below this limit */
   int resign_count;
   int draw_count;

   /* various things having to do with time management */
   chess_clock_t clock;

   int hash_found;
   int hash_cut;
   int branches_pruned;
   int branches_pruned_1st;
   int branches_pruned_lt3rd;

   /* Whether the engine is currently pondering or not, if it is, disable
    * time control.
    */
   bool pondering;
   move_t ponder_move;

   bool analysing;
   move_t analyse_move;
   board_t root_board;
   const char *analyse_fen;
   bool analyse_new;
   int analyse_undo;
   int analyse_moves_played;
   bool random_ok;

   /* Meta-data */
   char *name;

   /* Various function pointers, so we can easily customise things and
    * adjust to different UIs.
    */
   void (*output_iteration)(const char *, ...) __attribute__((format(printf, 1, 2)));
   void (*xboard_output)(const char *, ...) __attribute__((format(printf, 1, 2)));
   void (*error_output)(const char *, ...) __attribute__((format(printf, 1, 2)));

   bool (*check_keyboard)(struct game_t *game);
} game_t;

game_t *create_game(void);
void start_new_game(game_t *game);
void end_game(game_t *game);
void playmove(game_t *game, move_t move);
void play_null_move(game_t *game);
void takeback(game_t *game);
void takeback_null_move(game_t *game);
void backup_principle_variation(game_t *game, int depth, move_t move);
void truncate_principle_variation(game_t *game, int depth);
void setup_fen(game_t *game, const char *fen);
const char *make_fen_string(const game_t *game, char *buffer);

void set_transposition_table_size(game_t *game, size_t nelem);
void set_default_output_function(game_t *game, void (*func)(const char *, ...));
void set_xboard_output_function(game_t *game, void (*func)(const char *, ...));
void set_error_output_function(game_t *game, void (*func)(const char *, ...));

#endif
