#ifndef PIECES_H
#define PIECES_H
#include <stdbool.h>
#include <stdint.h>
#include "assert.h"

typedef int16_t eval_t;

typedef int piece_t;
typedef enum side_t {WHITE, BLACK, NONE, NUM_SIDES=NONE} side_t;

static inline side_t next_side(side_t side)
{
   assert(side == WHITE || side == BLACK);
   return side ^= BLACK;
}

#define MAX_PIECE_TYPES          24
#define MAX_PIECES_PER_SIDE      64

/* We define separate ranges in the piece list for the following categories of pieces:
 * - pawns (up to 12)
 * - royalty (up to 3, but normally only 1)
 * - sliders (up to 8, including promoted pieces)
 * - leapers (up to 24, including promoted pieces)
 * - mix/rest (rest, combined leapers/sliders)
 */
#define PIECE_LIST_PAWN_RANGE    12
#define PIECE_LIST_ROYAL_RANGE    3
#define PIECE_LIST_SLIDER_RANGE  12
#define PIECE_LIST_LEAPER_RANGE  20
#define PIECE_LIST_MIX_RANGE     16
#define PIECE_LIST_ROYAL_START   PIECE_LIST_PAWN_RANGE
#define PIECE_LIST_SLIDER_START  (PIECE_LIST_ROYAL_START + PIECE_LIST_ROYAL_RANGE)
#define PIECE_LIST_LEAPER_START  (PIECE_LIST_SLIDER_START + PIECE_LIST_SLIDER_RANGE)
#define PIECE_LIST_MIX_START     (PIECE_LIST_LEAPER_START + PIECE_LIST_LEAPER_RANGE)

#define PIECE_LIST_PAWN_MASK     ((1ull << PIECE_LIST_PAWN_RANGE)-1)
#define PIECE_LIST_ROYAL_MASK    (((1ull << (PIECE_LIST_ROYAL_START+PIECE_LIST_ROYAL_RANGE))-1)   & ~((1ull<<PIECE_LIST_ROYAL_START)-1))
#define PIECE_LIST_SLIDER_MASK   (((1ull << (PIECE_LIST_SLIDER_START+PIECE_LIST_SLIDER_RANGE))-1) & ~((1ull<<PIECE_LIST_SLIDER_START)-1))
#define PIECE_LIST_LEAPER_MASK   (((1ull << (PIECE_LIST_LEAPER_START+PIECE_LIST_LEAPER_RANGE))-1) & ~((1ull<<PIECE_LIST_LEAPER_START)-1))
#define PIECE_LIST_MIX_MASK      (((1ull << (PIECE_LIST_MIX_START+PIECE_LIST_MIX_RANGE))-1)       & ~((1ull<<PIECE_LIST_MIX_START)-1))

#define MAX_DIRECTIONS  12
/* To keep things simple, for symmetric pieces we only define the movement
 * tables for white.
 * These are then copied over by the initialisation code.
 */
typedef struct move_description_t {
   short int leap;         // Leaper step, may be 0 for pure sliders
   short int slide;        // Slider step, may be 0 for pure leapers
   short int count;        // Number of squares to slide
   short int skip;         // Skip=1 doesn't store the initial leaper step, skip>1 skips the first skip-1 slider steps
} move_description_t;

#define PF_SLIDE        0x01
#define PF_LEAP         0x02
#define PF_PROMOTE      0x04
typedef struct {
   const char *name, *symbol;    /* Name and symbol used in SAN moves */
   const char wfen, bfen;        /* FEN character, for white and black */
   const char *betza;            /* Betza move string */

   int value, phase;

   move_description_t move[MAX_DIRECTIONS];
   int flags;
   uint64_t piece_list_mask;     /* Mask in piece list for this piece type */
   int piece_list_base;          /* First element this piece type can occupy in the piece list */
} piece_description_t;
extern piece_description_t piece_info[MAX_PIECE_TYPES];
extern uint32_t attack_mask[512];
extern const int piece_types;

bool piece_attack_possible(int id, int from, int to);

#endif
