#include <stdint.h>
#include "bits64.h"
#include "score.h"
#include "eval.h"
#include "evalhash.h"
#include "hashkey.h"

/* Evaluation:
 *
 * Game phase
 * Material balance
 * Centralisation/advancement
 * Piece-king tropism (simplified king safety)
 * Passed pawn/advancement (value of a passer depends on the square it starts on)
 */

eval_t centre_table[MAILBOX_SIZE];
eval_t centre_sqr_table[MAILBOX_SIZE];
eval_t advance_table[MAILBOX_SIZE];
eval_t phase0 = 0;
eval_t tropism[512] = { 0 };

eval_t evaluate(game_t *game)
{
   const board_t *board = &game->board;
   int hash_ev = 0;
   eval_t ev = 0;
   eval_t eg = 0;
   int phase = 0;

   /* Try to save time by grabbing the evaluation score from a hash table */
   bool have_eval_hash = query_eval_table_entry(game->eval_hash, board->hash, &hash_ev);
#ifndef DEBUG_EVHASH
   if (have_eval_hash) {
      ev = hash_ev;
      goto exit;
   }
#endif

   for (side_t side = WHITE; side < NUM_SIDES; side++) {
      uint64_t mask = board->live_pieces[side];
      int oking = board->piece_list[PIECE_LIST_ROYAL_START + next_side(side) * MAX_PIECES_PER_SIDE].where;

      while (mask) {
         int idx = lsb64(mask) + side*MAX_PIECES_PER_SIDE;
         mask &= mask-1;

         int id = board->piece_list[idx].id;
         int rs = board->piece_list[idx].where;
         int sq = relative_square[side][rs];

         if (piece_info[id].value != CHECKMATE) {
            ev += piece_info[id].value;

            ev += centre_sqr_table[sq];

            /* King tropism evaluation */
            /* TODO: test if this is an improvement */
            //ev += piece_info[id].phase * (11 - tropism[256 + oking - rs]) / 2;

            eg += 2*advance_table[sq];
         } else {
            /* Keep the king away from the centre in the opening, but in
             * the centre in the end game.
             */
            ev -= centre_sqr_table[sq];
            eg += 2*centre_sqr_table[sq];
         }
         phase += piece_info[id].phase;

         /* Draw pieces close to a lone king */
         if (onebit64(board->live_pieces[next_side(side)]) && !(piece_info[id].flags & PF_PROMOTE))
            eg += (11 - tropism[256 + oking - rs]) * (11 - tropism[256 + oking - rs]);

         /* Push pawns in the end game */
         if (piece_info[id].flags & PF_PROMOTE) {
            int pid = promotion_map[sq % MAILBOX_FILES];
            ev += piece_info[pid].value / 100;
            eg += (advance_table[sq] - 3) * (advance_table[sq] - 3) / 2;
         }
      }

      eg = -eg;
      ev = -ev;
   }

   ev += (phase0 - phase) * eg/phase0;

#ifdef DEBUG_EVHASH
   /* Sanity check: the hashed score should equal the current score.
    * We only ever get here if we're debugging the evaluation hash.
    */
   if (!(!have_eval_hash || (hash_ev == ev))) {
      printf("%d %d %d %d\n", have_eval_hash, hash_ev, ev, symmetric);
      printf("0x%016llx\n", game->board.hash);
      print_bitboards(&game->board);
      printf("%s\n", make_fen_string(game, NULL));
   }
   assert(!have_eval_hash || (hash_ev == ev));
#endif

   /* Store the results of the evaluation in the evaluation hash table.
    * If the evaluation is symmetric, we can store it for the other side
    * to move as well, which will safe an evaluation after null-move.
    */
   store_eval_hash_entry(game->eval_hash, board->hash, ev);
   store_eval_hash_entry(game->eval_hash, get_stm_hash_key()^board->hash, ev);

exit:
   return (board->side_to_move == WHITE) ? ev : -ev;
}
