#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "game.h"
#include "eval.h"
#include "hashkey.h"

static void printfstderr(const char *msg, ...)
{
   static char buf[65536];
   va_list ap;
   va_start(ap, msg);
   vsnprintf(buf, 65535, msg, ap);
   va_end(ap);

   fprintf(stderr, "%s", buf);
}

static void (*default_iteration_output)(const char *, ...) = NULL;
static void (*default_xboard_output)(const char *, ...) = NULL;
static void (*default_error_output)(const char *, ...) = printfstderr;

game_t *create_game(void)
{
   game_t *game = calloc(1, sizeof *game);
   game->default_hash_size = HASH_TABLE_SIZE;
   return game;
}

/* Initialize the game datastructure to correspond to a new game */
void start_new_game(game_t *game)
{
   game->max_moves = 2000;
   free(game->ui);
   free(game->in_check);
   free(game->move_list);
   free(game->move_clock);
   free(game->score);
   game->ui             = calloc(game->max_moves, sizeof *game->ui);
   game->in_check       = calloc(game->max_moves, sizeof *game->in_check);
   game->move_list      = calloc(game->max_moves, sizeof *game->move_list);
   game->move_clock     = calloc(game->max_moves, sizeof *game->move_clock);
   game->score          = calloc(game->max_moves, sizeof *game->score);

   game->start_move_count = 0;
   memset(game->movelist, 0, sizeof game->movelist);
   memset(game->move_list, 0, game->max_moves * sizeof *game->move_list);
   memset(game->move_clock, 0, game->max_moves * sizeof *game->move_clock);
   memset(game->repetition_hash_table, 0, sizeof game->repetition_hash_table);
   game->moves_played = 0;

   memset(game->killer, 0, sizeof game->killer);
   memset(game->mate_killer, 0, sizeof game->mate_killer);
   memset(game->null_killer, 0, sizeof game->null_killer);
   memset(game->counter, 0, sizeof game->counter);
   memset(game->history, 0, sizeof game->history);
   game->max_history = 0;

   clear_board(&game->board);
   setup_fen(game, game->start_fen);

   /* Setup promotion choices */
   setup_promotions(&game->board);

   /* Initialise evaluation terms */
   phase0 = 0;
   for (side_t side = WHITE; side < NUM_SIDES; side++) {
      uint64_t mask = game->board.live_pieces[side];

      while (mask) {
         int idx = lsb64(mask) + side*MAX_PIECES_PER_SIDE;
         mask &= mask-1;

         int id = game->board.piece_list[idx].id;
         phase0 += piece_info[id].phase;
      }
   }

   if (game->default_hash_size == 0) game->default_hash_size = HASH_TABLE_SIZE;

   set_transposition_table_size(game, game->default_hash_size);

   game->analysing = false;
   game->pondering = false;
   game->random_ok = true;
   game->ponder_move = 0;

   game->draw_count = 0;
   game->resign_count = 0;

   game->output_iteration = default_iteration_output;
   game->xboard_output = default_xboard_output;
   game->error_output = default_error_output;
}

/* Frees the memory used by a game */
void end_game(game_t *game)
{
   if (game) {
      destroy_hash_table(game->transposition_table);
      destroy_eval_hash_table(game->eval_hash);
      free(game->start_fen);
      free(game->xb_setup);
      free(game->in_check);
      free(game->score);
      free(game->ui);
      free(game->move_list);
      free(game->move_clock);
      free(game);
   }
}

void set_transposition_table_size(game_t *game, size_t nelem)
{
   /* Round to the next-lowest power of 2 */
   nelem |= (nelem >> 1);
   nelem |= (nelem >> 2);
   nelem |= (nelem >> 4);
   nelem |= (nelem >> 8);
   nelem |= (nelem >> 16);
   nelem |= (nelem >> 32);
   nelem >>= 1;
   nelem++;

   game->default_hash_size = nelem;
   destroy_hash_table(game->transposition_table);
   destroy_eval_hash_table(game->eval_hash);
   game->transposition_table = NULL;
   game->eval_hash = NULL;
   game->transposition_table = create_hash_table(game->default_hash_size);
   game->eval_hash = create_eval_hash_table(game->default_hash_size/16);
}

void set_default_output_function(game_t *game, void (*func)(const char *, ...))
{
   game->output_iteration = default_iteration_output = func;
}

void set_xboard_output_function(game_t *game, void (*func)(const char *, ...))
{
   game->xboard_output = default_xboard_output = func;
}

void set_error_output_function(game_t *game, void (*func)(const char *, ...))
{
   game->error_output = default_error_output = func;
}


void playmove(game_t *game, move_t move)
{
   bool irreversible_move = is_irreversible_move(&game->board, move);

   makemove(&game->board, move, game->ui + game->moves_played);

   game->move_list[game->moves_played] = move;
   game->repetition_hash_table[game->board.hash&0xFFFF]++;
   game->moves_played++;

   /* Test whether the move is a checking move */
   //collect_attack_info(&game->board);
   //game->in_check[game->moves_played] = game->board.in_check;
   //assert(!player_in_check(&game->board, next_side[game->board.side_to_move]));
   //assert(game->board.in_check == player_in_check(&game->board, game->board.side_to_move));

   /* Advance the 50-move counter if this move was reversible, reset if it
    * wasn't.
    * FIXME: need to factor in moves by non-reversible pieces (pawns).
    */
   game->board.fifty_counter = irreversible_move ? 0
                                                 : (game->board.fifty_counter + 1);
}

void play_null_move(game_t *game)
{
   //makemove(&game->board, 0, game->ui + game->moves_played);
   game->board.side_to_move = next_side(game->board.side_to_move);

   game->repetition_hash_table[game->board.hash&0xFFFF]++;
   game->move_list[game->moves_played] = 0;
   game->moves_played++;
   //game->in_check[game->moves_played] = game->board.in_check;
   game->board.hash ^= get_stm_hash_key();
}

void takeback(game_t *game)
{
   game->moves_played--;
   game->repetition_hash_table[game->board.hash&0xFFFF]--;
   move_t move = game->move_list[game->moves_played];
   unmakemove(&game->board, move, game->ui + game->moves_played);
}

void takeback_null_move(game_t *game)
{
   game->moves_played--;
   game->repetition_hash_table[game->board.hash&0xFFFF]--;
   game->board.side_to_move = next_side(game->board.side_to_move);
   game->board.hash ^= get_stm_hash_key();
}

/************************************* 
 * Propagate the principle variation *
 *************************************/
void backup_principle_variation(game_t *game, int depth, move_t move)
{
   int c;

   /* Store move */
   if (move) {
      game->length_of_variation[depth] = depth;
      game->principle_variation[depth][depth] = move;
      //printf("Store at %d %s\n", depth, move_string(move, NULL));
   }

   /* Copy tree below last stored move */
   if (game->length_of_variation[depth+1] > depth) {
      for (c=depth+1; c<game->length_of_variation[depth+1]; c++) {
         game->principle_variation[c][depth] = game->principle_variation[c][depth+1];
      }
      game->length_of_variation[depth] = game->length_of_variation[depth+1];
   }
}

void truncate_principle_variation(game_t *game, int depth)
{
   /* Truncate the principle variation at this depth */
   game->length_of_variation[depth] = depth;
}

void setup_fen(game_t *game, const char *fen)
{
   board_t *board = &game->board;
   const char *s = fen;
   int square = FILES*(RANKS-1);
   int prev_rank = 2*FILES;
   int n;

   clear_board(board);

   while(*s && (*s != ' ') && (*s != '[') && square>=0) {
      switch (*s) {
         case '/':
            square -= prev_rank;
            break;
         case '*':
            square++;
            break;
         case '1':
            if (isdigit(s[1])) {
               square += 10;
               break;
            }
         case '2':
         case '3':
         case '4':
         case '5':
         case '6':
         case '7':
         case '8':
         case '9':
         case '0':
            square += (*s - '1')+1;
            break;

         default:
            (void)0;
            side_t side = NONE;
            int piece = -1;
            for (int n=0; n<piece_types; n++) {
               /* Determine piece colour */
               side_t match_side = NONE;
               if (piece_info[n].wfen == *s) match_side = WHITE;
               if (piece_info[n].bfen == *s) match_side = BLACK;
               if (match_side == NONE) continue;

               side = match_side;
               piece = n;
            }
            if (piece == -1) {
               printf("Error: unknown piece type '%c' (bad FEN %s)\n", *s, fen);
               return;
            }
            add_piece_to_list(board, side, piece, square_to_mailbox[square]);
            square++;
            break;
      }
      s++;
   }

   while (*s && isspace(*s)) s++;

   /* Side to move */
   set_side_to_move(board, (*s == 'b') ? BLACK : WHITE);

   while (*s && !isspace(*s)) s++;
   while (*s && isspace(*s)) s++;

   /* Castle rights */
   while (*s && !isspace(*s)) {
      if (isupper(*s)) board->piece_list[PIECE_LIST_ROYAL_START].virgin = true;
      if (islower(*s)) board->piece_list[PIECE_LIST_ROYAL_START+MAX_PIECES_PER_SIDE].virgin = true;
      s++;
   }

   while (*s && isspace(*s)) s++;

   /* Half-move clock */
   while(*s && (*s == ' ')) s++;
   n = 0;
   sscanf(s, "%d\n", &n);
   board->fifty_counter = n;
   while(*s && (*s != ' ')) s++;

   /* Move counter */
   while(*s && (*s == ' ')) s++;
   n = 0;
   sscanf(s, "%d\n", &n);
   game->start_move_count = 2*((n > 1) ? (n-1) : 0);
   while(*s && (*s != ' ')) s++;

   /* Transfer piece-list to board */
   board->hash = (board->side_to_move) ? 0ull : get_stm_hash_key();
   for (side_t side=WHITE; side<=BLACK; side++) {
      uint64_t mask = board->live_pieces[side];
      while (mask) {
         int n = lsb64(mask);
         mask &= mask-1;
         int index = n + MAX_PIECES_PER_SIDE*side;
         board->mailbox[board->piece_list[index].where] = 1 + index;
         board->hash ^= get_hash_key(side, board->piece_list[index].id, board->piece_list[index].where);
      }
   }

   game->repetition_hash_table[board->hash&0xFFFF] = 1;
}


const char *make_fen_string(const game_t *game, char *buffer)
{
   static char static_buffer[4096];
   char *fen = buffer;
   int n = 0;
   int r, f;

   if (!fen) fen = static_buffer;
   fen[0] = '\0';

   /* First record: board position */
   /* Scan all ranks */
   for (r = RANKS-1; r>=0; r--) {
      int count = 0;
      for (f = 0; f < FILES; f++) {
         int square = f + r*FILES;
         int msqr   = square_to_mailbox[square];

         if (game->board.mailbox[msqr] == 0xFE) {
            if (count) n += snprintf(fen+n, 4096 - n, "%d", count);
            count = 0;
            n += snprintf(fen+n, 4096-n, "*");
            continue;
         }

         /* Empty? */
         if (!game->board.mailbox[msqr]) {
            count++;
            continue;
         }

         /* Not empty, do we have a count? */
         if (count) n += snprintf(fen+n, 4096 - n, "%d", count);
         count = 0;

         /* Print piece */
         side_t side = piece_colour_on_square(&game->board, msqr);
         int piece_id = piece_id_on_square(&game->board, msqr);

         n += snprintf(fen+n, 4096-n, "%c", (side == WHITE) ?  piece_info[piece_id].wfen : piece_info[piece_id].bfen );
      }
      if (count) n += snprintf(fen+n, 4096 - n, "%d", count);
      if (r) n += snprintf(fen+n, 4096 - n, "/");
   }

   n += snprintf(fen+n, 4096 - n, " ");

   /* Second record: side to move */
   if (game->board.side_to_move == WHITE)
      n += snprintf(fen+n, 4096 - n, "w ");
   else
      n += snprintf(fen+n, 4096 - n, "b ");

   /* Third record: castling rights */
   bool wrote_castle = false;
   for (side_t side = WHITE; side <= BLACK; side++) {
      if (game->board.piece_list[PIECE_LIST_ROYAL_START + side * MAX_PIECES_PER_SIDE].virgin) {
         const char *cstr = "gG";
         n += snprintf(fen+n, 4096 - n, "%c", cstr[side]);
         wrote_castle = true;
      }
   }
   if (!wrote_castle) n += snprintf(fen+n, 4096 - n, "-");
   n += snprintf(fen+n, 4096 - n, " ");

   /* Fifth and sixth record: half-move counter and full-move counter */
   int par = (game->moves_played&1) && next_side(game->board.side_to_move);
   n += snprintf(fen+n, 4096 - n, "%d ", game->board.fifty_counter);
   n += snprintf(fen+n, 4096 - n, "%d", (int)(game->start_move_count + game->moves_played + par)/2 + 1);

   return fen;
}
