#include <string.h>
#include "hashtable.h"
#include "score.h"

#define HASH_BUCKETS    4

static inline size_t map_key_to_index(uint64_t key, size_t nelem)
{
   return key & ( nelem - 1);
}

static inline uint32_t map_key_to_lock(uint64_t key)
{
   return key >> 32;
}

hash_table_t *create_hash_table(size_t nelem)
{
   hash_table_t *table;
   if (!nelem) return NULL;

   table = calloc(1, sizeof *table);
   table->data = calloc(nelem + HASH_BUCKETS, sizeof *table->data);
   table->number_of_elements = nelem;
   return table;
}

void destroy_hash_table(hash_table_t *table)
{
   if (table) {
      free(table->data);
      free(table);
   }
}

/* Encrypt transposition table entry using the "xor trick" for lockless
 * hashing.
 */
static void crypt(hash_table_entry_t *hash)
{
   uint64_t *h;
   assert(sizeof(hash_table_entry_t) == 2*sizeof(uint64_t));
   h = (uint64_t *)hash;
   h[0] ^= h[1];
}

//static lock_t hash_lock = 0;
void store_table_entry(hash_table_t *table, uint64_t key, int depth_lower, int score_lower, int depth_upper, int score_upper, move_t move)
{
   hash_table_entry_t *data;
   hash_table_entry_t *worst_data = NULL;
   size_t index, b;
   uint32_t lock;

   if (!table)
      return;

   /* Map the key onto the array index, check if entry is there */
   index = map_key_to_index(key, table->number_of_elements);
   lock = map_key_to_lock(key);
   worst_data = data = table->data + index;

   //acquire_lock(&hash_lock);

   /* Check all buckets */
   for (b=0; b<HASH_BUCKETS; b++) {
      hash_table_entry_t hash = data[b];
      crypt(&hash);
      if (hash.lock == lock || hash.generation == 0) {
         worst_data = data+b;
         break;
      }

      if (hash.depth_lower < worst_data->depth_lower ||
          hash.depth_upper < worst_data->depth_upper ||
          hash.generation < table->generation) {
         worst_data = data+b;
      }
   }
   data = worst_data;

   data->lock = lock;
   data->depth_lower = depth_lower;
   data->score_lower = score_lower;
   data->depth_upper = depth_upper;
   data->score_upper = score_upper;
   data->move        = move;
   data->generation = table->generation;
   crypt(data);
   table->write_count++;
   //release_lock(&hash_lock);
}

bool retrieve_table(hash_table_t *table, uint64_t key, int *depth_lower, int *score_lower, int *depth_upper, int *score_upper, move_t *move)
{
   size_t index, b;
   uint32_t lock;

   if (!table)
      return false;

   /* Map the key onto the array index, check if entry is there */
   index = map_key_to_index(key, table->number_of_elements);
   lock = map_key_to_lock(key);

   *depth_lower = -1;
   *depth_upper = -1;
   *score_lower = -CHECKMATE;
   *score_upper =  CHECKMATE;
   *move        =  0;

   //acquire_lock(&hash_lock);

   /* Check all buckets */
   b = 0;
   for (b = 0; b<HASH_BUCKETS; b++) {
      hash_table_entry_t hash = table->data[index+b];
      crypt(&hash);
      if (hash.lock == lock) {
         hash.generation = table->generation;

         *depth_lower = hash.depth_lower;
         *depth_upper = hash.depth_upper;
         if (hash.depth_lower)
            *score_lower = hash.score_lower;
         if (hash.depth_upper)
            *score_upper = hash.score_upper;
         *move        = hash.move;
         //release_lock(&hash_lock);
         return true;
      }
   }

   //release_lock(&hash_lock);
   return false;
}

void prepare_hashtable_search(hash_table_t *table)
{
   table->generation++;
   table->write_count = 0;
}

int count_unused_table_entries(hash_table_t *table)
{
   size_t n;
   int count;
   if (!table)
      return 0;

   count = 0;

   for (n=0; n<table->number_of_elements; n++) {
      if (table->data[n].generation == 0)
         count++;
   }

   return count;
}

