#include <stdbool.h>
#include "movegen.h"
#include "movelist.h"
#include "board.h"
#include "move_string.h"

bool movegen(movelist_t *ml, const board_t *board)
{
   bool legal = true;
   clear_movelist(ml);

   const piece_t *mailbox = board->mailbox;

   side_t oside = next_side(get_side_to_move(board));
   int oking = board->piece_list[oside*MAX_PIECES_PER_SIDE + PIECE_LIST_ROYAL_START].where;

   int base = MAX_PIECES_PER_SIDE*get_side_to_move(board);
   uint64_t mask = board->live_pieces[get_side_to_move(board)];
   uint64_t lm;
   
   /* Pure leapers */
   lm = mask & (PIECE_LIST_LEAPER_MASK | PIECE_LIST_ROYAL_MASK);
   mask ^= lm;
   while (lm && legal) {
      int pidx = lsb64(lm) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= lm-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      for (int n=0; n<MAX_DIRECTIONS && legal; n++) {
         move_description_t md = piece_info[id].move[n];
         assert(md.slide == 0);
         if (md.leap == 0) break;

         int to = from + md.leap;

         if (mailbox[to] == B) continue;
         if (mailbox[to] && piece_colour_on_square(board, to) == get_side_to_move(board)) continue;
         legal = legal && (to != oking);
         push_movelist(ml, encode_move(from, to));
      }
   }

   /* Pure sliders */
   lm = mask & PIECE_LIST_SLIDER_MASK;
   mask ^= lm;
   while (lm && legal) {
      int pidx = lsb64(lm) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= lm-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      for (int n=0; n<MAX_DIRECTIONS && legal; n++) {
         move_description_t md = piece_info[id].move[n];
         assert(md.leap == 0);
         if (md.slide == 0) break;

         int to = from + md.slide;

         if (mailbox[to] == B) continue;

         /* Slider steps */
         int s = 0;
         for ( ; s<md.skip-1 && mailbox[to] == 0; s++) to += md.slide;
         for ( ; s<md.count  && mailbox[to] == 0; s++) {
            push_movelist(ml, encode_move(from, to));
            to += md.slide;
         }
         if (mailbox[to] == B) continue;
         if (mailbox[to] && piece_colour_on_square(board, to) == get_side_to_move(board)) continue;
         legal = legal && (to != oking);
         push_movelist(ml, encode_move(from, to));
      }
   }

   /* Mixed pieces */
   lm = mask & PIECE_LIST_MIX_MASK;
   mask ^= lm;
   while (lm && legal) {
      int pidx = lsb64(lm) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= lm-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      for (int n=0; n<MAX_DIRECTIONS && legal; n++) {
         move_description_t md = piece_info[id].move[n];
         if (md.leap == 0 && md.slide == 0) break;

         int to = from + md.leap;

         if (mailbox[to] == B) continue;

         /* Slider steps */
         if (mailbox[to] == 0 && md.slide) {
            /* Store leaper step */
            if (md.leap && !md.skip)
               push_movelist(ml, encode_move(from, to));

            int s = 0;
            to += md.slide;
            for ( ; s<md.skip-1 && mailbox[to] == 0; s++) to += md.slide;
            for ( ; s<md.count  && mailbox[to] == 0; s++) {
               push_movelist(ml, encode_move(from, to));
               to += md.slide;
            }
            if (mailbox[to] == B) continue;
         } else if (md.slide && md.skip) {
            continue;
         }
         if (mailbox[to] && piece_colour_on_square(board, to) == get_side_to_move(board)) continue;
         legal = legal && (to != oking);
         push_movelist(ml, encode_move(from, to));
      }
   }

   /* Pawn moves and captures */
   lm = mask & PIECE_LIST_PAWN_MASK;
   mask ^= lm;
   while (lm && legal) {
      int pidx = lsb64(lm) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= lm-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      move_description_t md = piece_info[id].move[0];
      int to = from + md.leap;
      if (mailbox[to] == 0)
         push_movelist(ml, encode_move(from, to));

      for (int d=-1; d<=1; d+=2) {
         move_description_t md = piece_info[id].move[0];
         int to = from + md.leap + d;

         if (mailbox[to] == B) continue;
         if (mailbox[to] == 0 || piece_colour_on_square(board, to) == get_side_to_move(board)) continue;
         legal = legal && (to != oking);
         push_movelist(ml, encode_move(from, to));
      }
   }

   /* We should have covered all pieces now */
   assert(mask == 0);

   /* King castling/double jumps */
   mask = board->live_pieces[get_side_to_move(board)] & PIECE_LIST_ROYAL_MASK;
   while (mask && legal) {
      int pidx = lsb64(mask) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      mask &= mask-1;

      if (!board->piece_list[pidx].virgin) continue;

      /* Castling move is stored in next entry */
      id++;
      for (int n=0; n<MAX_DIRECTIONS; n++) {
         move_description_t md = piece_info[id].move[n];
         if (md.leap == 0 && md.slide == 0) break;

         int to = from + md.leap;

         /* Should be non-capture */
         if (mailbox[to]) continue;

         //if (mailbox[to] && piece_colour_on_square(board, to) == get_side_to_move(board)) continue;
         legal = legal && (to != oking);
         push_movelist(ml, encode_move(from, to));
      }
   }

   return legal;
}


bool capturegen(movelist_t *ml, const board_t *board)
{
   bool legal = true;
   clear_movelist(ml);

   const piece_t *mailbox = board->mailbox;

   side_t oside = next_side(get_side_to_move(board));
   int oking = board->piece_list[oside*MAX_PIECES_PER_SIDE + PIECE_LIST_ROYAL_START].where;

   int base = MAX_PIECES_PER_SIDE*get_side_to_move(board);
   uint64_t mask = board->live_pieces[get_side_to_move(board)];
   uint64_t lm;
   
   /* Pure leapers */
   lm = mask & (PIECE_LIST_LEAPER_MASK | PIECE_LIST_ROYAL_MASK);
   mask ^= lm;
   while (lm && legal) {
      int pidx = lsb64(lm) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= lm-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      for (int n=0; n<MAX_DIRECTIONS && legal; n++) {
         move_description_t md = piece_info[id].move[n];
         assert(md.slide == 0);
         if (md.leap == 0) break;

         int to = from + md.leap;

         if (mailbox[to] == B || mailbox[to] == 0) continue;
         if (mailbox[to] && piece_colour_on_square(board, to) == get_side_to_move(board)) continue;
         legal = legal && (to != oking);
         push_movelist(ml, encode_move(from, to));
      }
   }

   /* Pure sliders */
   lm = mask & PIECE_LIST_SLIDER_MASK;
   mask ^= lm;
   while (lm && legal) {
      int pidx = lsb64(lm) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= lm-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      for (int n=0; n<MAX_DIRECTIONS && legal; n++) {
         move_description_t md = piece_info[id].move[n];
         assert(md.leap == 0);
         if (md.slide == 0) break;

         int to = from + md.slide;

         if (mailbox[to] == B) continue;

         /* Slider steps */
         int s = 0;
         for ( ; s<md.skip-1 && mailbox[to] == 0; s++) to += md.slide;
         for ( ; s<md.count  && mailbox[to] == 0; s++) to += md.slide;
         if (mailbox[to] == B || mailbox[to] == 0) continue;
         if (mailbox[to] && piece_colour_on_square(board, to) == get_side_to_move(board)) continue;
         legal = legal && (to != oking);
         push_movelist(ml, encode_move(from, to));
      }
   }

   /* Mixed pieces */
   lm = mask & PIECE_LIST_MIX_MASK;
   mask ^= lm;
   while (lm && legal) {
      int pidx = lsb64(lm) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= lm-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      for (int n=0; n<MAX_DIRECTIONS && legal; n++) {
         move_description_t md = piece_info[id].move[n];
         if (md.leap == 0 && md.slide == 0) break;

         int to = from + md.leap;

         if (mailbox[to] == B) continue;

         /* Slider steps */
         if (mailbox[to] == 0 && md.slide) {
            int s = 0;
            to += md.slide;
            for ( ; s<md.skip-1 && mailbox[to] == 0; s++) to += md.slide;
            for ( ; s<md.count  && mailbox[to] == 0; s++) to += md.slide;
            if (mailbox[to] == B) continue;
         } else if (md.slide && md.skip) {
            continue;
         }
         if (mailbox[to] == 0) continue;
         if (mailbox[to] && piece_colour_on_square(board, to) == get_side_to_move(board)) continue;
         legal = legal && (to != oking);
         push_movelist(ml, encode_move(from, to));
      }
   }

   /* Pawn moves and captures */
   lm = mask & PIECE_LIST_PAWN_MASK;
   mask ^= lm;
   while (lm && legal) {
      int pidx = lsb64(lm) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= lm-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      move_description_t md = piece_info[id].move[0];

      for (int d=-1; d<=1; d+=2) {
         int to = from + md.leap + d;

         if (mailbox[to] == B) continue;
         if (mailbox[to] == 0 || piece_colour_on_square(board, to) == get_side_to_move(board)) continue;
         legal = legal && (to != oking);
         push_movelist(ml, encode_move(from, to));
      }
   }

   /* We should have covered all pieces now */
   assert(mask == 0);

   return legal;
}

bool calculate_piece_can_attack(const board_t *board, piece_t id, int from, int square)
{
   const piece_t *mailbox = board->mailbox;

   for (int n=0; n<MAX_DIRECTIONS; n++) {
      move_description_t md = piece_info[id].move[n];
      if (md.leap == 0 && md.slide == 0) break;

      int to = from + md.leap;

      // Pawns 
      if (piece_info[id].flags & PF_PROMOTE) {
         return abs(square - to) == 1;
      }

      if (mailbox[to] == B) continue;
      if (to == square) return true;

      /* Slider steps */
      if (mailbox[to] != B && md.slide) {
         int s = 0;
         to += md.slide;
         for ( ; s<md.skip-1 && mailbox[to] != B; s++) to += md.slide;
         for ( ; s<md.count  && mailbox[to] != B && to != square; s++) to += md.slide;
         if (mailbox[to] == B) continue;
         if (to == square) return true;
      }
      if (mailbox[to] == B) continue;
      if (piece_info[id].flags & PF_PROMOTE) {  // Pawns
         return abs(square - to) == 1;
      }
      if (to == square) return true;
   }

   return false;
}

bool in_check(const board_t *board, side_t side)
{
   bool check = false;
   side_t oside = next_side(side);

   const piece_t *mailbox = board->mailbox;
   uint64_t mask = board->live_pieces[oside];
   int base = MAX_PIECES_PER_SIDE*oside;
   int king = board->piece_list[PIECE_LIST_ROYAL_START + side*MAX_PIECES_PER_SIDE].where;

   /* Test leapers */
   uint64_t lm = mask & (PIECE_LIST_LEAPER_MASK|PIECE_LIST_ROYAL_MASK);
   mask ^= lm;
   while (lm && !check) {
      int pidx = lsb64(lm) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= lm-1;

      /* Test if attacks are possible at all.
       * This test has no false positives for leapers.
       */
      check = check || piece_attack_possible(id, from, king);
   }

   /* Test sliders */
   lm = mask & PIECE_LIST_SLIDER_MASK;
   mask ^= lm;
   while (lm && !check) {
      int pidx = lsb64(lm) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= lm-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      /* Test if attacks are possible at all.
       * This test may give a false positive (and of course the ray may be
       * blocked as well), but if it says no attack is possible, the test
       * is accurate.
       */
      if (!piece_attack_possible(id, from, king)) continue;

      for (int n=0; n<MAX_DIRECTIONS && !check; n++) {
         move_description_t md = piece_info[id].move[n];
         if (md.slide == 0) break;

         int to = from + md.slide;

         /* Slider steps, if useful */
         if (mailbox[to] == 0) {
            int s = 0;
            for ( ; s<md.skip-1 && mailbox[to] == 0; s++) to += md.slide;
            for ( ; s<md.count  && mailbox[to] == 0; s++) {
               to += md.slide;
            }
            if (mailbox[to] == B) continue;
         }
         check = check || (to == king);
      }
   }

   /* Test mixed pieces */
   lm = mask & PIECE_LIST_MIX_MASK;
   mask ^= lm;
   while (lm && !check) {
      int pidx = lsb64(lm & ~PIECE_LIST_PAWN_MASK) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      lm &= (lm & ~PIECE_LIST_PAWN_MASK)-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      /* Test if attacks are possible at all.
       * This test may give a false positive (and of course the ray may be
       * blocked as well), but if it says no attack is possible, the test
       * is accurate.
       */
      if (!piece_attack_possible(id, from, king)) continue;

      for (int n=0; n<MAX_DIRECTIONS && !check; n++) {
         move_description_t md = piece_info[id].move[n];
         if (md.leap == 0 && md.slide == 0) break;

         int to = from + md.leap;

         check = check || (to == king);

         if (mailbox[to] == B || check) continue;

         /* Slider steps, if useful */
         if (mailbox[to] == 0 && md.slide) {
            int s = 0;
            to += md.slide;
            for ( ; s<md.skip-1 && mailbox[to] == 0; s++) to += md.slide;
            for ( ; s<md.count  && mailbox[to] == 0; s++) {
               to += md.slide;
            }
            if (mailbox[to] == B) continue;
         }
         check = check || (to == king);
      }
   }

   /* Scan pawns */
   while (mask && !check) {
      int pidx = lsb64(mask) + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      mask &= mask-1;

      /* Test if attacks are possible at all.
       * This test has no false positives for leapers.
       */
      check = check || piece_attack_possible(id, from, king);
   }

   return check;
}

bool legal_movegen(movelist_t *movelist, board_t *board)
{
   bool ret = movegen(movelist, board);
   if (!ret) return ret;

   /* Filter out illegal moves */
   side_t me = board->side_to_move;
   for (int n = 0; n<movelist->num_moves; n++) {
      unmake_info_t u;

      makemove(board, movelist->move[n], &u);
      bool check = in_check(board, me);
      unmakemove(board, movelist->move[n], &u);

      if (!check) continue;
      movelist->move[n] = movelist->move[movelist->num_moves-1];
      movelist->num_moves--;
      n--;
   }

   return ret;
}

uint64_t get_attackers(const board_t *board, int square, side_t side)
{
   uint64_t attackers = 0;
   uint64_t mask = board->live_pieces[side] & ~(PIECE_LIST_SLIDER_MASK | PIECE_LIST_MIX_MASK);
   int base = MAX_PIECES_PER_SIDE*side;
   const piece_t *mailbox = board->mailbox;

   /* Begin with leapers */
   while (mask) {
      int bit  = lsb64(mask);
      int pidx = bit + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      mask &= mask-1;

      if (piece_attack_possible(id, from, square))
         attackers |= (1ull << bit);
   }

   /* Sliders */
   mask = board->live_pieces[side] & PIECE_LIST_SLIDER_MASK;
   while (mask) {
      int bit  = lsb64(mask);
      int pidx = bit + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      mask &= mask-1;

      if (!piece_attack_possible(id, from, square)) continue;

      for (int n=0; n<MAX_DIRECTIONS; n++) {
         move_description_t md = piece_info[id].move[n];
         if (md.slide == 0) break;

         int to = from + md.slide;

         if (mailbox[to] == 0) {
            int s = 0;
            for ( ; s<md.skip-1 && mailbox[to] == 0; s++) to += md.slide;
            for ( ; s<md.count  && mailbox[to] == 0 && to != square; s++) {
               to += md.slide;
            }
            if (mailbox[to] == B) continue;
         }
         if (to == square) {
            attackers |= (1ull << bit);
            break;
         }
      }
   }

   /* Mixed pieces */
   mask = board->live_pieces[side] & PIECE_LIST_MIX_MASK;
   while (mask) {
      int bit  = lsb64(mask);
      int pidx = bit + base;
      int from = board->piece_list[pidx].where;
      int id   = board->piece_list[pidx].id;
      mask &= mask-1;

      /* Captured pieces should be skipped, but these should already not be
       * part of the bitmask of live pieces.
       */
      assert(from != 0);

      /* Test if attacks are possible at all.
       * This test may give a false positive (and of course the ray may be
       * blocked as well), but if it says no attack is possible, the test
       * is accurate.
       */
      if (!piece_attack_possible(id, from, square)) continue;

      for (int n=0; n<MAX_DIRECTIONS; n++) {
         move_description_t md = piece_info[id].move[n];
         if (md.leap == 0 && md.slide == 0) break;

         int to = from + md.leap;

         if (mailbox[to] == B);

         /* Slider steps, if useful */
         if ((mailbox[to] == 0 || from == to) && md.slide && to != square) {
            int s = 0;
            to += md.slide;
            for ( ; s<md.skip-1 && mailbox[to] == 0; s++) to += md.slide;
            for ( ; s<md.count  && mailbox[to] == 0 && to != square; s++) {
               to += md.slide;
            }
            if (mailbox[to] == B) continue;
         }
         if (to == square) {
            attackers |= (1ull << bit);
            break;
         }
      }
   }

   return attackers;
}
