#include "movesort.h"
#include "score.h"
#include "eval.h"
#include "move_string.h"
#include "see.h"

extern void print_board(const board_t *board);

#define HISTORY_HEURISTIC
#define SORT_SEE

#define in_check(n) (game->in_check[game->moves_played + n])

static bool moves_are_equal(move_t m1, move_t m2) { return m1 == m2; }

static bool is_killer(const game_t *game, int depth, move_t prev_move, move_t move) __attribute__((unused));
static bool is_killer(const game_t *game, int depth, move_t prev_move, move_t move)
{
   if (moves_are_equal(game->killer[0][depth], move) ||
       moves_are_equal(game->killer[1][depth], move))
      return true;
   return false;
}

/* Store moves that kill a branch in the killer slots, but only if:
 *  - we were not in check at the beginning of this move
 *  - the move is not a promotion (already high in the tree)
 *  - the move was not a capture (already high in the tree)
 */
void store_killer(game_t *game, move_t move, int depth)
{
   if (in_check(0) || is_promotion_move(&game->board, move) || is_capture_move(&game->board, move))
      return;

   if (moves_are_equal(move, game->killer[0][depth])) {
      /* The move was the first killer - do nothing */
   } else {
      /* This was either the last killer (out of 2 or 3), or it's a new
       * move. Either way, Degrade first killer to second killer (etc) and
       * store the new first killer.
       */
      game->killer[1][depth]=game->killer[0][depth];
      game->killer[0][depth]=move;
   }
}


void store_mate_killer(game_t *game, move_t move, int depth, int best_score)
{
   if (best_score >= (CHECKMATE-1000))
      game->mate_killer[depth] = move;
}

void store_null_killer(game_t *game, int depth, move_t move)
{
   if (!is_capture_move(&game->board, move))
      game->null_killer[depth] = move;
}

void store_counter_move(game_t *game, move_t move, move_t omove)
{
   if (in_check(0) || is_promotion_move(&game->board, move) || is_capture_move(&game->board, move))
      return;

   side_t side = game->board.side_to_move;
   int    from = get_move_from(omove);
   int    to   = get_move_to(omove);
   game->counter[from][to][side] = move;
}

bool is_counter_move(const game_t *game, move_t move, move_t omove)
{
   if (in_check(0) || is_promotion_move(&game->board, move) || is_capture_move(&game->board, move))
      return false;

   side_t side = game->board.side_to_move;
   int    from = get_move_from(omove);
   int    to   = get_move_to(omove);
   return game->counter[from][to][side] == move;
}

void adjust_history_score(game_t *game, move_t move, int score)
{
#ifdef HISTORY_HEURISTIC
   if (is_capture_move(&game->board, move) ||
      is_promotion_move(&game->board, move) || in_check(0))
      return;

   side_t side = game->board.side_to_move;
   int piece = get_move_piece(&game->board, move);
   int to = get_move_to(move);

   game->history[side][piece][to] += score;
   if (abs(game->history[side][piece][to]) > game->max_history)
      game->max_history = abs(game->history[side][piece][to]);

   /* Rescale as needed */
   if (game->max_history > 32767) {
      game->max_history /= 2;
      for (side = WHITE; side <= BLACK; side++)
         for (piece=0; piece<piece_types; piece++)
            for (to=0; to<256; to++)
               game->history[side][piece][to] /= 2;
   }
#endif
}

int get_move_history(const game_t *game, move_t move)
{
#ifdef HISTORY_HEURISTIC
   if (is_capture_move(&game->board, move) ||
      is_promotion_move(&game->board, move) || in_check(0))
      return 0;

   side_t side = game->board.side_to_move;
   int piece = get_move_piece(&game->board, move);
   int to = get_move_to(move);

   return game->history[side][piece][to];
#else
   return 0;
#endif
}

void score_moves(const game_t *game, movelist_t *movelist, int depth, const move_t hash_move)
{
   const board_t *board = &game->board;
   int n;

   int *score = movelist->score;

   for (n=0; n<movelist->num_moves; n++) {
      move_t move = movelist->move[n];
      int from = get_move_from(move);
      int to = get_move_to(move);

      score[n] = 0;

      /* Give the highest priority to the move from the hash table, if it
       * exists.
       */
      if (moves_are_equal(hash_move, move)) {
         score[n] += CHECKMATE;
         continue;
      }

      /* TODO: advances of free pawn, advances of candidate pawns */

      /* Mate killer: a killer move that produced a mate score */
      if (moves_are_equal(game->mate_killer[depth], move)) {
         score[n] += CHECKMATE-100;
         continue;
      }

      /* Regular captures
       * The score is 1100+the SSE score, which sorts winning captures below
       * promotion moves (with a score >= 1150) and losing captures with a
       * score <=1000 below everything else.
       */
      if (is_capture_move(board, move)) {
#ifdef SORT_SEE
         //if (!print) {
         //   print_board(&game->board);
         //   print = true;
         //}
         int see_score = see(board, move);
         //printf("# n=%d, %s = %d (%d, %d)\n", n, lan_move_string(&game->board, move, NULL), score[n], piece_info[victim].value, see_score);
         if (see_score > 0)
            score[n] += 1300+see_score;
         else
            score[n] += see_score;
#else
         int piece  = piece_id_on_square(board, from);
         int victim = piece_id_on_square(board, to);

         score[n] += 1300 + (2*piece_info[victim].value - piece_info[piece].value)/8;
#endif

         continue;
      }

      /* Promotion moves */
      if (is_promotion_move(board, move)) {
         int to = get_move_to(move);
         int id = promotion_map[to % MAILBOX_FILES];

         score[n] += 1300+piece_info[id].value/2;
         continue;
      }

      /* Award killers a higher score, so they are searched first
       * Given the weights here, winning captures are normally searched
       * before killer moves, losing captures are searched after killer
       * moves.
       */
      if (moves_are_equal(game->killer[0][depth], move)) {
         score[n] += 1300;
         continue;
      }
      if (moves_are_equal(game->killer[1][depth], move)) {
         score[n] += 1290;
         continue;
      }

      /* Null killer: a move that failed high after a null move */
      if (moves_are_equal(game->null_killer[depth], move)) {
         score[n] += 1200;
         continue;
      }

      /* Also look at killers from earlier positions. This doesn't make a
       * big difference, but it speeds things up very slightly still and
       * isn't likely to hurt.
       */
      if (depth>2 && moves_are_equal(game->killer[0][depth-2], move)) {
         score[n] += 1280;
         continue;
      }
      if (depth>2 && moves_are_equal(game->killer[1][depth-2], move)) {
         score[n] += 1270;
         continue;
      }

      if (game->max_history) {
         score[n] = 100 + 500 * get_move_history(game, move) / game->max_history;
         if (score[n] == 100) score[n] = 0;
      }

      /* If we don't know what else to do, order moves based on the
       * centralisation table.
       */
      if (score[n] == 0)
         score[n] += centre_table[to] - centre_table[from];

      /* Idea: pawn advances are typically bad, so penalise pawn pushes
       * (except maybe to the 7th rank) and sort them lower down.
       */
   }
}

