#include <stdarg.h>
#include <stdbool.h>
#include "bits64.h"
#include "genrand.h"
#include "move_string.h"
#include "movegen.h"
#include "movelist.h"
#include "search.h"
#include "board.h"
#include "play.h"
#include "score.h"

#define out if (game->output_iteration) game->output_iteration
#define xb if (game->xboard_output) game->xboard_output

bool show_fail_low  = true;
bool show_fail_high = true;

static int min(int x, int y)
{
   return (x<y)?x:y;
}

static int max(int x, int y)
{
   return (x>y)?x:y;
}

static void show(const game_t *game, const char *msg, ...)
{
   static char buf[65536];
   va_list ap;
   va_start(ap, msg);
   vsnprintf(buf, 65535, msg, ap);
   va_end(ap);

   xb("%s", buf);
   out("%s", buf);
}


static void print_principle_variation(game_t *game)
{
   int c;
   int root_move_number = 0;

   if (!game->output_iteration) return;

   out("   ");
   if (game->board.side_to_move) {
      root_move_number = 1;
      out("%d. ... ", (int)(game->start_move_count + game->moves_played)/2+1);
   }
   for (c=0; c<game->length_of_variation[0]; c++) {
      if (((root_move_number+c) & 1) == 0)
         out("%d. ", (int)(game->start_move_count + root_move_number+game->moves_played+c)/2+1);

      move_t move = game->principle_variation[c][0];
      out("%-5s ", short_move_string(&game->board, move, NULL));
      playmove(game, move);
   }
   for (c=0; c<game->length_of_variation[0]; c++)
      takeback(game);
}


static void print_principle_variation_xb(game_t *game)
{
   int c;
   int root_move_number = 0;

   xb("   ");
   if (game->board.side_to_move) {
      root_move_number = 1;
      xb("%d. ... ", (int)(game->start_move_count + game->moves_played)/2+1);
   }
   for (c=0; c<game->length_of_variation[0]; c++) {
      if (((root_move_number+c) & 1) == 0)
         xb("%d. ", (int)(game->start_move_count + root_move_number+game->moves_played)/2+1);

      move_t move = game->principle_variation[c][0];
      xb("%-5s ", short_move_string(&game->board, move, NULL));
      playmove(game, move);
   }
   for (c=0; c<game->length_of_variation[0]; c++)
      takeback(game);
}

play_state_t think(game_t *game, int max_depth)
{
   play_state_t state = SEARCH_OK;
   uint64_t start_time = get_timer();
   bool mate_search = false;
   int depth;

   /* Start the clock */
   start_clock(&game->clock);

   game->clock.extra_time = 0;
   game->clock.root_moves_played = game->moves_played;
   game->clock.nodes_searched = 0;

   game->hash_found = 0;
   game->hash_cut = 0;
   game->branches_pruned = 0;
   game->branches_pruned_1st = 0;
   game->branches_pruned_lt3rd = 0;
   abort_search = false;
   deepest_q = 0;

   for (side_t side = WHITE; side<=BLACK; side++)
      if (onebit64(game->board.live_pieces[side])) mate_search = true;

   movelist_t *movelist = game->movelist+0;
   legal_movegen(movelist, &game->board);

   xb("# Position '%s'\n", make_fen_string(game, NULL));

   /* Prepare the transpostion table for a new search iteration:
    * Resets the write count and increases the generation counter.
    */
   prepare_hashtable_search(game->transposition_table);
   game->board.check = in_check(&game->board, game->board.side_to_move);

   if (movelist->num_moves == 0) {
      if (game->board.check)
         return SEARCH_GAME_ENDED_MATE;
      else
         return SEARCH_GAME_ENDED_STALEMATE;
   }

   int e = 0;
   game->clock.extra_time = game->clock.time_left;
   game->best_move[0] = 0;
   int score = search(game, -CHECKMATE, CHECKMATE, 1, 0);
   game->clock.extra_time = 0;
   move_t move = game->best_move[0];

   /* Iterative deepening loop */
   for (depth=2; depth<=max_depth && !abort_search; depth++) {
      //scale_history();
      //movelist.rewind();

      for (int window = 50; window < 2*CHECKMATE; window *= 3) {
         if (depth == 2) score = 0;
         if (depth <= 4) window = CHECKMATE / (depth - 1);
         if (mate_search && !is_mate_score(score)) window = 2*CHECKMATE-1;
         int alpha = max(score - window, -CHECKMATE);
         int beta  = min(score + window,  CHECKMATE);
         if (mate_search) beta = CHECKMATE;

         int new_score = search(game, alpha, beta, depth+e, 0);
         game->clock.extra_time = 0;

         if (abort_search) break;
         score = new_score;

         if (score > alpha && score < beta) {
            move = game->principle_variation[0][0];
            break;
         }

         if (is_mate_score(score)) mate_search = true;

         /* Report a fail-low */
         if (score < alpha && move) {
            move_t move = game->best_move[0];
            if (show_fail_low) {
               out("% 3d.  %6.2f  %9d %+6.2f  ", depth, (get_timer()-start_time)/1000000.0, (int)game->clock.nodes_searched, score/100.0);
               out("   %d.", (int)(game->start_move_count + game->moves_played)/2+1);
               if (game->board.side_to_move) out(" ...");
               out(" %s?", short_move_string(&game->board, move, NULL));
               out("\n");

               xb("% 3d % 5d %6d %9d ", depth, score, (int)(get_timer()-start_time)/10000, (int)game->clock.nodes_searched);
               xb("   %d.", (int)(game->start_move_count + game->moves_played)/2+1);
               if (game->board.side_to_move) xb(" ...");
               xb(" %s?", short_move_string(&game->board, move, NULL));
               xb("\n");
            }
            game->clock.extra_time = (7*game->clock.time_left)/10;
         }


         /* Report a fail-high */
         if (score >= beta) {
            move = game->principle_variation[0][0];
            if (show_fail_high) {
               out("% 3d.  %6.2f  %9d %+6.2f  ", depth, (get_timer()-start_time)/1000000.0, (int)game->clock.nodes_searched, score/100.0);
               out("   %d.", (int)(game->start_move_count + game->moves_played)/2+1);
               if (game->board.side_to_move) out(" ...");
               out(" %s!", short_move_string(&game->board, move, NULL));
               out("\n");

               xb("% 3d % 5d %6d %9d ", depth, score, (int)(get_timer()-start_time)/10000, (int)game->clock.nodes_searched);
               xb("   %d.", (int)(game->start_move_count + game->moves_played)/2+1);
               if (game->board.side_to_move) xb(" ...");
               xb(" %s!", short_move_string(&game->board, move, NULL));
               xb("\n");
            }
            game->clock.extra_time = (5*game->clock.time_left)/10;
         }
      }
      /* Convert score to centi-pawn */
      if (!is_mate_score(score)) score = score*100/piece_info[0].value;
      out("% 3d.  %6.2f  %9d %+6.2f  ", depth, (get_timer() - start_time)/1000000.0, (int)game->clock.nodes_searched, score/100.0);
      print_principle_variation(game);
      out("\n");

      xb("% 3d % 5d %6d %9d ", depth, score, (int)(get_timer()-start_time)/10000, (int)game->clock.nodes_searched);
      print_principle_variation_xb(game);
      xb("\n");

      if (is_mate_score(score) && (CHECKMATE - abs(score)) == game->length_of_variation[0] && move && !game->analysing) break;
      if (abort_search) break;
      if (game->movelist[0].num_moves == 1 && move && !game->pondering && !game->analysing) break;
      if (game->analysing && game->check_keyboard && game->check_keyboard(game)) break;

      /* Check if we have enough time for the next iteration, assuming an
       * effective branching ratio of 2.
       */
      uint64_t time_limit = get_chess_clock_time_for_move(&game->clock)/2;
      if ( game->clock.check_clock && peek_timer(&game->clock) > time_limit && move)
         break;

      /* Update tablebase, if needed.
       * Don't do this if the root position is a candidate TB position,
       * since we don't arrive here unless we ran out of time constructing
       * the tablebase.
       */
   }
   xb("# Position '%s'\n", make_fen_string(game, NULL));
   show(game, "# Nodes: %zd (%.2e NPS) Cut-offs: %d Cut1: %d (%.2f%%%%) Cut<4: %d (%.2f%%%%), %d in hash (%.2f%%%% cut)\n",
      game->clock.nodes_searched,
      game->clock.nodes_searched * 1000000.0 / (get_timer()+1 - start_time),
      game->branches_pruned,
      game->branches_pruned_1st, 100.*game->branches_pruned_1st/game->branches_pruned,
      game->branches_pruned_lt3rd, 100.*game->branches_pruned_lt3rd/game->branches_pruned,
      game->hash_found, 100.*game->hash_cut/game->hash_found);
   show(game, "# DQ: %d\n", deepest_q);

   while (game->analysing && !abort_search) {
      if (game->check_keyboard && game->check_keyboard(game)) abort_search |= 1;
   }

   if (score < game->resign_threshold)
      game->resign_count++;
   else
      game->resign_count = 0;

   if (abs(score) < game->draw_threshold)
      game->draw_count++;
   else
      game->draw_count = 0;

   if (game->moves_played > 400)
      game->draw_count += 50;

   assert(move);
   if (!move)
      move = movelist->move[genrandui() % movelist->num_moves];
   playmove(game, move);

   return state;
}

bool analyse(game_t *game)
{
   move_t ponder_move = game->ponder_move;

   if (game->analyse_move) return false;

   game->analyse_move = 0;
   game->root_board = game->board;

   /* Search the current position */
   abort_search = false;
   while (game->analysing && !abort_search) {
      if (think(game, MAX_SEARCH_DEPTH) == SEARCH_OK)
         takeback(game);
      else {
         while (game->analysing && !abort_search) {
            if (game->check_keyboard && game->check_keyboard(game)) abort_search |= 1;
         }
      }
   }
   if (game->analyse_move) playmove(game, game->analyse_move);
   game->analyse_move = 0;

   game->ponder_move = ponder_move;

   return true;
}


