#include "assert.h"
#include "search.h"
#include "score.h"
#include "movegen.h"
#include "eval.h"
#include "movesort.h"
#include "hashtable.h"
#include "move_string.h"
#include "see.h"

/* Function call to check how much time we have left
 * We do the check every time clock_positions nodes have been searched.
 * Don't do this during QS - there aren't that many QS nodes anyway.
 */
int deepest_q;
static const int clock_positions = 0x00007FFF;
#define check_clock() \
   if (game->clock.max_nodes && (game->clock.nodes_searched >= game->clock.max_nodes)) abort_search |= 1;\
   if (game->check_keyboard && game->check_keyboard(game)) abort_search |= 1;\
   if (game->clock.check_clock && ((game->clock.nodes_searched & clock_positions)==0))\
      abort_search |= game->clock.check_clock(&game->clock)

volatile bool abort_search;

static inline int min(int x, int y)
{
   return (x<y) ? x : y;
}

static inline int max(int x, int y)
{
   return (x>y) ? x : y;
}

/* Return the number of repetitions of the current position in the game
 * tree.
 */
int count_repetition(const game_t *game)
{
   int n;
   int count;
   count = 0;

   for (n=game->moves_played-1; n>=0; n--) {
      count += (game->ui[n].hash == game->board.hash);
   }

   return count;
}

static inline bool position_repeated_once(const game_t *game)
{
   return game->repetition_hash_table[game->board.hash&0xFFFF]>=2 && count_repetition(game)>=1;
}

static inline bool position_repeated(const game_t *game)
{
   return game->repetition_hash_table[game->board.hash&0xFFFF]>=2 && count_repetition(game)>=2;
}

static inline bool legal_draw(const game_t *game)
{
   return game->board.fifty_counter > 101 ||
          position_repeated(game);
}

int get_extension(move_t move, int move_score)
{
   /* Check extension: only for safe checks.
    * Check extension is handled elsewhere
    */
   return 0;
}

int get_reduction(move_t move, int move_score, int move_count, int draft)
{
   int r = 0;

   if (move_score > 1000) return 0;

   r = 1;
   if (r && draft > 4) r += draft / 4;

   return r;
}

bool zugzwang_threat(const board_t *board)
{
   if ((board->live_pieces[board->side_to_move] & ~(PIECE_LIST_ROYAL_MASK | PIECE_LIST_PAWN_MASK)))
      return true;

   return false;
}

int qsearch(game_t *game, int alpha, int beta, int draft, int ply)
{
   int static_score = evaluate(game);

   if (draft < deepest_q) deepest_q = draft;
   if (ply >= MAX_TOTAL_DEPTH) return static_score; 

   if (static_score > alpha) alpha = static_score;
   if (alpha >= beta) return alpha;

   check_clock();
   if (abort_search) return 0;

   truncate_principle_variation(game, ply);
   movelist_t *movelist = game->movelist + ply;

   bool legal = capturegen(movelist, &game->board);
   if (!legal) return -ILLEGAL;

   /* Score the moves */
   move_t best_move = 0, move = 0;
   score_moves(game, movelist, ply, best_move);

   int best_score = static_score;
   int score = 0;
   int moves_searched = 0;
   while (movelist->cur_move < movelist->num_moves) {
      if (abort_search) return 0;
      move = next_move(movelist);
      if (!is_capture_move(&game->board, move)) continue;
      if (see(&game->board, move) < 0) continue;
      game->best_move[ply] = move;
      playmove(game, move);
      game->clock.nodes_searched++;
      score = -qsearch(game, -beta, -alpha, draft-1, ply+1);
      takeback(game);

      if (score == ILLEGAL)
         continue;
      moves_searched++;
      if (score > best_score) {
         best_score = score;
         best_move = move;
      }
      break;
   }

   if (best_score > alpha) alpha = best_score;
   while (alpha < beta && movelist->cur_move < movelist->num_moves) {
      move = next_move(movelist);

      if (!is_capture_move(&game->board, move) && !is_promotion_move(&game->board, move)) continue;

      playmove(game, move);
      game->clock.nodes_searched++;

      score = -qsearch(game, -(alpha+1), -alpha, draft-1, ply+1);

      /* Verification search, only useful when the window is not a null
       * window.
       */
      if (score > alpha && score < beta)
         score = -qsearch(game, -beta, -alpha, draft-1, ply+1);

      takeback(game);

      if (score == ILLEGAL) continue;
      moves_searched++;

      if (score > best_score) {
         best_score = score;
         best_move = game->best_move[ply] = move;
      }

      if (score > alpha) {    /* New best line found */
         alpha = score;
         best_move = move;

         /* Update and store principle variation */
         backup_principle_variation(game, ply, move);
      }
   }

   if (alpha >= beta) {        /* Beta cutoff */
      /* Store good moves in the killer slots, but only if:
       *  - we were not in check at the beginning of this move
       *  - the move is not a promotion (already high in the list)
       *  - the move was not a capture (already high in the list)
       */
      game->branches_pruned_1st += (moves_searched == 1);
      game->branches_pruned_lt3rd += (moves_searched < 3);
      game->branches_pruned++;
   }

   /* Abort if we've exceeded our allowed time to think */
   if (abort_search) return 0;

   return best_score;
}

int search(game_t *game, int alpha, int beta, int draft, int ply)
{
   assert(game);
   assert(beta > alpha);

   /* Truncate the principle variation at this depth here, otherwise we may
    * end up with rubbish at the end of the PV.
    */
   truncate_principle_variation(game, ply);

   /* The very first thing we do is to check if this position has occurred
    * three times in the game tree, or whether no irreversible move has
    * been played in 50 moves (100 ply). Either of these conditions means
    * we should flag a draw and return immediately.
    * During the search, we simply score positions as a draw if they have
    * occurred at least once before. From the top level, we don't claim a
    * draw until three repetitions have been made.
    */
   if (ply > 0) {
      if (legal_draw(game) || position_repeated_once(game))
         return 0;
   }

   /* Check whether our search time for this move has expired */
   check_clock();
   if (abort_search) return 0;

   /* Check whether we're looking for a checkmate.
    * If so, prune branches that are worse than the best mate found so far
    * (mate distance pruning).
    */
   alpha = max(alpha, -CHECKMATE + ply);
   beta  = min(beta,   CHECKMATE - ply - 1);
   if (alpha >= beta)
      return alpha;

   if (draft <= 0 || ply > MAX_SEARCH_DEPTH)
      return qsearch(game, alpha, beta, draft, ply);

   /* Look up this position in the transposition table */
   move_t hash_move = 0;
   int    hash_score_upper = 0;
   int    hash_score_lower = 0;
   int    hash_depth_upper = 0;
   int    hash_depth_lower = 0;
   bool   have_hash = retrieve_table(game->transposition_table, game->board.hash, &hash_depth_lower, &hash_score_lower, &hash_depth_upper, &hash_score_upper, &hash_move);
   game->hash_found += have_hash;

   /* Internal iterative deepening */
   if (!have_hash && alpha != beta-1 && draft > 1) {
      search(game, alpha, beta, draft-1, ply);
      have_hash = retrieve_table(game->transposition_table, game->board.hash, &hash_depth_lower, &hash_score_lower, &hash_depth_upper, &hash_score_upper, &hash_move);
   }

   if (have_hash && game->board.fifty_counter < 80) {
      if (hash_depth_lower>0) hash_score_lower = score_from_hashtable(hash_score_lower, ply);
      if (hash_depth_upper>0) hash_score_upper = score_from_hashtable(hash_score_upper, ply);

      if ((hash_depth_lower >= draft || is_mate_score(hash_score_lower)) && hash_depth_lower>0 && hash_score_lower >= beta) {
         game->hash_cut++;
         return hash_score_lower;
      }

      if ((hash_depth_upper >= draft || is_mate_score(hash_score_upper)) && hash_depth_upper>0 && hash_score_upper <= alpha) {
         game->hash_cut++;
         return hash_score_upper;
      }

      if (hash_score_upper == hash_score_lower && alpha != beta-1 && hash_depth_upper>0 && (hash_depth_upper >= draft || is_mate_score(hash_score_upper))) {
         game->hash_cut++;
         return hash_score_upper;
      }
   }

   /* First try a null-move.
    * This serves a double-purpose: it is also used as a check-detection.
    */
   bool check = game->board.check;
   int static_score = (check) ? -CHECKMATE
                              : evaluate(game);

   move_t prev_move = game->moves_played ?  game->move_list[game->moves_played-1] : 0;
   bool avoid_null = beta > alpha+1 ||
                     check ||
                     draft < 4 ||
                     prev_move == 0 ||
                     zugzwang_threat(&game->board) ||
                     static_score < beta ||
                     is_mate_score(beta);
   if (!avoid_null) {
      int r = 2 + draft / 4;
      int score;
      play_null_move(game);
      game->clock.nodes_searched++;
      score = -search(game, -beta, -(beta-1), draft - 1 - r, ply+1);
      takeback_null_move(game);
      if (score >= beta && !is_mate_score(score) && !abort_search)
         return score;
   }

   /* Futility pruning */
   if (!check && draft == 1 && ((static_score + 700) < alpha || (static_score - 700) > beta) && !is_mate_score(alpha))
      return static_score;

   /* Generate moves - not needed in the root, where it was already done by
    * the iterative deepening loop.
    * If the other side is in check, we return a -ILLEGAL score which is
    * caught one level up to mark the preceding move as illegal.
    */
   movelist_t *movelist = game->movelist + ply;
   bool legal = true;
   if (ply == 0)
      rewind_movelist(movelist);
   else
      legal = movegen(movelist, &game->board);
   if (!legal) return -ILLEGAL;

   int legal_moves = movelist->num_moves;

   /* Score the moves */
   score_moves(game, movelist, ply, hash_move);

   /* Search the first move */
   int score = -CHECKMATE;
   move_t move = 0;
   while (legal_moves) {
      move = next_move(movelist);
      game->best_move[ply] = move;
      playmove(game, move);
      game->clock.nodes_searched++;
      game->board.check = in_check(&game->board, game->board.side_to_move);
      int e = draft <= 1 && game->board.check && see(&game->board, move) > 0;
      score = -search(game, -beta, -alpha, draft-1 + e, ply+1);
      takeback(game);

      if (score == ILLEGAL) {
         legal_moves--;
         continue;
      }
      break;
   }

   /* If there are no legal moves, the game is over */
   if (legal_moves == 0) {
      score = check ? (-CHECKMATE + ply) : STALEMATE;
      return score;
   }

   if (score > alpha) {           /* New best line found */
      alpha = score;

      adjust_history_score(game, move, draft * draft);

      /* Update and store principle variation */
      backup_principle_variation(game, ply, move);
   }

   /* Search remaining moves; we know there is at least one legal move at
    * this point, so don't bother tracking those anymore.
    */
   move_t best_move = move;
   int moves_searched = 1;
   int best_score = score;
   while (alpha < beta && movelist->cur_move < movelist->num_moves) {
      move = next_move(movelist);

      /* Futility pruning */
      if (draft == 1 && !check && !is_mate_score(beta) && static_score < beta) {
         /* Losing captures */
         if (is_capture_move(&game->board, move) && get_move_score(movelist) < 1000) continue;

         /* Unsafe quiet moves */
         //if (!is_capture_move(&game->board, move) && see(&game->board, move) < 0) continue;
      }

      playmove(game, move);
      game->clock.nodes_searched++;

      int e = get_extension(move, get_move_score(movelist));
      int r = (e == 0) ? get_reduction(move, get_move_score(movelist), moves_searched, draft) : 0;
      //if (in_check(&game->board, next_side(game->board.side_to_move))) {
      //   takeback(game);
      //   legal_moves--;
      //   continue;
      //}
      /* Check extension: extend checking moves */
      game->board.check = in_check(&game->board, game->board.side_to_move);
      if (game->board.check) {
         e = 1;
         r = 0;
      }
      score = -search(game, -(alpha+1), -alpha, draft-1 + e - r, ply+1);

      /* Verification search, only useful when the window is not a null
       * window.
       */
      if (score > alpha && score < beta)
         score = -search(game, -beta, -alpha, draft-1 + e, ply+1);

      takeback(game);

      if (score == ILLEGAL) continue;

      if (score > best_score) {
         best_score = score;
         best_move = game->best_move[ply] = move;
      }

      if (score > alpha) {    /* New best line found */
         alpha = score;
         best_move = move;

         adjust_history_score(game, move, draft * draft);

         /* Update and store principle variation */
         backup_principle_variation(game, ply, move);
      }
      moves_searched++;
   }

   if (alpha >= beta) {        /* Beta cutoff */
      /* Store good moves in the killer slots, but only if:
       *  - we were not in check at the beginning of this move
       *  - the move is not a promotion (already high in the list)
       *  - the move was not a capture (already high in the list)
       */
      store_killer(game, best_move, ply);
      store_mate_killer(game, best_move, ply, best_score);
      if (prev_move == 0)
         store_null_killer(game, ply, best_move);

      game->branches_pruned_1st += (moves_searched == 1);
      game->branches_pruned_lt3rd += (moves_searched < 3);
      game->branches_pruned++;

      for (int n=0; n<movelist->cur_move-1; n++)
         adjust_history_score(game, movelist->move[n], -draft * draft);
   }


   /* Abort if we've exceeded our allowed time to think */
   if (abort_search) return 0;

   /* Store best move for retrieval from one level up in the search (for IID) */
   game->best_move[ply] = best_move;

   /* store evaluation in the transposition table */
   if (!have_hash) {
      hash_move = 0;
      hash_score_upper = CHECKMATE;
      hash_score_lower = -CHECKMATE;
      hash_depth_upper = 0;
      hash_depth_lower = 0;
   }

   if (best_score < beta) {
      hash_score_upper = best_score;
      hash_depth_upper = draft;
      hash_move = 0;
   }
   if (best_score > alpha) {
      hash_score_lower = best_score;
      hash_depth_lower = draft;
      hash_move = best_move;
   } 
   if (hash_depth_lower>0) hash_score_lower = score_to_hashtable(hash_score_lower, ply);
   if (hash_depth_upper>0) hash_score_upper = score_to_hashtable(hash_score_upper, ply);
   store_table_entry(game->transposition_table, game->board.hash, hash_depth_lower, hash_score_lower, hash_depth_upper, hash_score_upper, hash_move);


   return best_score;
}

