#include <string.h>
#include "bits64.h"
#include "pieces.h"
#include "score.h"
#include "board.h"
#include "movegen.h"
#include "see.h"

#undef PRINT_SEE_DEBUG_MESSAGES

static inline int max(int x, int y)
{
   return (x>y)?x:y;
}

static inline int find_least_valued_piece(const board_t *board, side_t side, uint64_t own)
{
   uint64_t mask = own & board->live_pieces[side];
   int lvp = -1;
   int value = CHECKMATE+1;

   while (mask) {
      int index = lsb64(mask) + side*MAX_PIECES_PER_SIDE;
      int id    = board->piece_list[index].id;
      mask &= mask-1;

      if (piece_info[id].value < value) {
         lvp = index;
         value = piece_info[id].value;
      }
   }

   return lvp;
}

static int move_value(const board_t *board, move_t move)
{
   if (is_capture_move(board, move)) {
      int ptaken = piece_id_on_square(board, get_move_to(move));
      int score  = piece_info[ptaken].value;

      return score;
   }

   return 0;
}

/* Perform static exchange evaluation for a given capture move.
 * Uses the "swap" algorithm
 * (http://chessprogramming.wikispaces.com/SEE+-+The+Swap+Algorithm).
 * returns the evaluation in centipawns.
 */
int see(const board_t *board, const move_t move)
{
   int score[256];
   int depth;
   int side = board->side_to_move;
   int square = get_move_to(move);
   int last_piece = get_move_piece(board, move);
   uint64_t attackers[NUM_SIDES];

   /* Initial gain: the piece on the target square */
   depth = 0;
   score[depth] = move_value(board, move);

   /* Easy: if the piece we use to capture the enemy piece is worth less
    * than the piece we captured, it's always a winning capture, so return a
    * rough estimate for the SEE score.
    */
   if (score[depth] > piece_info[last_piece].value)
      return score[depth];

   attackers[WHITE] = get_attackers(board, square, WHITE);
   attackers[BLACK] = get_attackers(board, square, BLACK);

#if 0
   uint64_t mask = attackers[WHITE];
   printf("WHITE: ");
   while (mask) {
      int idx = lsb64(mask);
      mask &= mask-1;
      printf("%s ", square_str[board->piece_list[idx].where]);
   }
   printf("\n");

   printf("BLACK: ");
   mask = attackers[BLACK];
   while (mask) {
      int idx = lsb64(mask) + MAX_PIECES_PER_SIDE;
      mask &= mask-1;
      printf("%s ", square_str[board->piece_list[idx].where]);
   }
   printf("\n");

   printf("SEE: %2d %d %s%s % 4d 0x%016llx\n", depth, side, square_str[get_move_from(move)], square_str[square], score[depth], attackers[side]);
#endif

   int index = board->mailbox[get_move_from(move)]-1;
   if (index >= 0)
      attackers[side] ^= 1ull << (index & (MAX_PIECES_PER_SIDE - 1));

   if (attackers[next_side(side)]) {
      do {
         /* Consider next move - one ply deeper, flip side */
         depth++;
         side = next_side(side);
         assert(depth < 256);
         /* Update score (speculative) */
         score[depth] = piece_info[last_piece].value - score[depth-1];
         /* Perform LVA capture */
         if (attackers[side]) {
            /* Find the least valued attacker */
            int index = find_least_valued_piece(board, side, attackers[side]);
            assert(index >= 0);
            last_piece = board->piece_list[index].id;
            //printf("     %2d %d %-3s  % 4d 0x%016llx\n", depth, side, square_str[board->piece_list[index].where], score[depth], attackers[side]);

            /* Update the list of attackers/defenders */
            attackers[side] ^= 1ull << (index % MAX_PIECES_PER_SIDE);
         }
      } while(attackers[board->side_to_move]);

      /* Compute final score */
      depth++;
      while (--depth) {
         score[depth-1] = -max(-score[depth-1], score[depth]);
      }
   }
   return score[depth];
}


