/* Playing Tamerlane Chess and Grande Acedrex requires some special
 * considerations from a chess program.
 *
 * For the rules, see
 * http://history.chess.free.fr/acedrex.htm
 *
 * The board size needs to be 12x12 for Grande Acedrex and 11x10 for
 * Tamerlane (with two extra squares, so 13x10 really).
 *
 * Grande Acedrex has the following pieces:
 * 
 * King      - as K
 * Gryphon   - as F and then optionally as R in same quadrant as F move
 * Crocodile - as B
 * Giraffe   - as (2,3) leaper (Zebra Z)
 * Unicorn   - as N and then otionally as B but only forward and outward
 * Lion      - as (0,3) or (1,3) leaper
 * Rook      - as R
 * Pawn      - as P, promote to file piece
 * (8 types)
 *
 * Tamerlane has the following pieces:
 *
 * King      - as K
 * Prince    - as K
 * Adv. king - as K
 * Ferz      - as F
 * Wazir     - as W
 * Giraffe   - as F, then straight after skipping 3 squares
 * Picket    - as nA, then optional outward as B
 * Knight    - N
 * Rook      - R
 * Elephant  - A
 * Camel     - (1,3) leaper
 * Dababa    - D
 * Pawn      - as P; promote according to initial file (so really 11 pawns)
 * (23 types)
 *
 * King, prince and adventitious king are different: only the king may
 * occupy the enemy citadel and only the adventitious king can occupy the
 * own citadel.
 *
 * For Courier Chess we would need:
 *
 * King      - as K
 * Man       - as K
 * Ferz      - as F
 * Wazir     - as W
 * Courier   - as B
 * Knight    - N
 * Rook      - R
 * Elephant  - A
 * Pawn      - as P; promote to Ferz
 * (9 types)
 *
 * Needs 23 piece types -> 5 bits
 * Colour bit
 * Initial/unmoved bit
 * Sentinel/off-board bit
 * Together this fits into a normal byte.
 *
 * The board could be stored in a (3+3+12) ranks x (3+12) files = 18x15
 * board.
 *
 * All pieces have a two-step (multi-step) move:
 *  1. A leaper move (offset for second step)
 *  2. A slider move (with a skip and repeat count)
 * We need to decide whether the second leg is optional or not...
 * Store separate black/white move options
 */
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <stdint.h>
#include <stdio.h>
#include <stdbool.h>
#include <ctype.h>
#include <time.h>
#include "assert.h"
#include "genrand.h"
#include "hashkey.h"
#include "bits64.h"
#include "move.h"
#include "movelist.h"
#include "board.h"
#include "game.h"
#include "xstring.h"
#include "timer.h"
#include "movegen.h"
#include "play.h"
#include "move_string.h"
#include "search.h"
#include "score.h"
#include "eval.h"
#include "keypressed.h"

#ifdef __APPLE__
#define __unix__
#endif
#ifdef __unix__
#include <unistd.h>
#include <signal.h>
#endif

static int  draw_count = 0;
static int  resign_count = 0;
static eval_t draw_threshold = 0;
static eval_t resign_threshold = -1200;
static int  tc_moves = 40;
static int  tc_time  = 60000;
static int  tc_inc   = 0;
static bool ansi = true;
static bool xboard48 = false;
static char deferred[256];

static FILE *log_f = NULL;
static char buf[65536];

//const char *piece_to_char = "P.CR.............A..U...G..LKp.cr.............a..u...g..lk";
//const char *piece_to_char = "P.BR........Z....A.LUKp.br........z....a.luk";
//const char *grand_acedrex_fen = "rluzbakbzulr/12/12/pppppppppppp/12/12/12/12/PPPPPPPPPPPP/12/12/RLUZBAKBZULR w Kk 0 1";
const char *grand_acedrex_fen = "rlugcakcgulr/12/12/pppppppppppp/12/12/12/12/PPPPPPPPPPPP/12/12/RLUGCAKCGULR w Gg 0 1";
const char *grand_acedrex_no_castle_fen = "rlugcakcgulr/12/12/pppppppppppp/12/12/12/12/PPPPPPPPPPPP/12/12/RLUGCAKCGULR w - 0 1";

void print_board(const board_t *board)
{
   FILE *file = stdout;

   // 46/45 works for purple/cyan
   // 43/41 works for red/yellow
   const char *bg_colour_string[] = { "\033[45m", "\033[43m", "\033[44m" };
   char mark[256];
   int colour[256];
   int pieces[256];
   bool occupied[256];
   side_t side[256];
   int c, n;

   for (int r=0; r<RANKS; r++) {
      for (int f=0; f<FILES; f++) {
         int square = f + r*FILES;

         occupied[square] = false;
         colour[square]   = 2;
         pieces[square]   = 0xff;
         mark[square]     = ' ';
      }
   }

   for (int r=0; r<RANKS; r++) {
      for (int f=0; f<FILES; f++) {
         int msqr = MAILBOX_FIRST + f + r*MAILBOX_FILES;
         int square = f + r*FILES;

         colour[square] = ((square / FILES) ^ (square % FILES)) & 1;

         if (board->mailbox[msqr]) {
            int pidx = board->mailbox[msqr] - 1;
            occupied[square] = true;
            side[square] = pidx >= MAX_PIECES_PER_SIDE;
            pieces[square] = board->piece_list[pidx].id;
         } else {
            occupied[square] = false;
         }
      }
   }

   if (file != stdout) ansi = false;

   for (c=RANKS-1; c>=0; c--) {
      if (ansi) fprintf(file, "%2d", 1+c);
      if (ansi) fprintf(file, "\033[1m");
      for (n=0; n<FILES; n++) {
         int square = n + c*FILES;
         int piece  = pieces[square];
         if (ansi) fprintf(file, "%s", bg_colour_string[colour[square]]);
         if (occupied[square]) {
            bool white = (side[square] == WHITE);
            if (ansi) fprintf(file, "%s", white ? "\033[37m" : "\033[30m");
            fprintf(file, " %c", white?piece_info[piece].wfen:piece_info[piece].bfen);
         } else {
            if (ansi)
               fprintf(file, " %c", mark[square]);
            else
               fprintf(file, "%c%c", colour[square] ? '+' : '.', mark[square]);
         }
      }
      if (ansi) fprintf(file, "\033[0m");
      if (get_side_to_move(board) == BLACK && c == RANKS-1) fprintf(file, "*");
      if (get_side_to_move(board) == WHITE && c == 0) fprintf(file, "*");
      fprintf(file, "\n");
   }
   if (ansi) {
      fprintf(file, "  ");
      for (n=0; n<FILES; n++)
         fprintf(file, "%c ", 'a' + n);
      fprintf(file, "\n");
   }
}

static uint64_t perft(game_t *game, int depth, int root)
{
   movelist_t *movelist = game->movelist + depth;
   side_t me = get_side_to_move(&game->board);
   uint64_t nodes = 0;
   int n;

   if (abort_search) return 0;
   if (depth == 0) return !in_check(&game->board, next_side(me));

   /* Check if previous move left the player in check */
   bool legal = movegen(movelist, &game->board);
   if (!legal) return 0;

   for (n=0; n<movelist->num_moves; n++) {
      uint64_t count = 0;
      playmove(game, movelist->move[n]);
      count = perft(game, depth-1, root - 1);
      nodes += count;
      if (root > 0)
         printf("%8s %10"PRIu64" %10"PRIu64"\n", lan_move_string(&game->board, movelist->move[n], NULL), count, nodes);
      takeback(game);
      if (abort_search) break;
   }
   return nodes;
}

void print_movelist(const movelist_t *ml, board_t *board)
{
   for (int n=0; n<ml->num_moves; n++) {
      printf("%3d/%3d ", n+1, ml->num_moves);
      move_t move    = ml->move[n];
      int from       = get_move_from(move);
      int to         = get_move_to(move);
      char token     = (board->mailbox[to] == 0) ? '-' : 'x';
      int id         = piece_id_on_square(board, from);
      const char *p  = "";
      if (is_promotion_move(board, move)) {
         int id = promotion_map[get_move_to(move) % MAILBOX_FILES];
         p = piece_info[id].symbol;
      }

      char str[20];
      snprintf(str, 20, "%s%s%c%s%s", piece_info[id].symbol, square_str[from], token, square_str[to], p);

      printf("%10s ", str);
      printf("%10s ", short_move_string(board, move, NULL));
      printf("%10s ", lan_move_string(board, move, NULL));
      printf("\n");
   }
}

#ifdef __unix__
static sig_t old_signal_handler;

void interrupt_computer(int i)
{
   abort_search = true;
   if (old_signal_handler)
      old_signal_handler(i);
}
#endif

static bool stdin_is_terminal(void)
{
#ifdef __unix__
   return isatty(fileno(stdin));
#else
   return true;
#endif
}

static bool stdout_is_terminal(void)
{
#ifdef __unix__
   return isatty(fileno(stdout));
#else
   return true;
#endif
}

#define TIME_BUFFER  200
static void set_time_from_string(game_t *game, const char *timestr)
{
   float milliseconds;
   sscanf(timestr, "%g", &milliseconds);
   milliseconds *= 10;

   /* Reserve a bit of time for when we reach the time control, so we don't get penalised by a time
    * loss. We don't need this for a Fischer clock.
    */
   if (milliseconds > TIME_BUFFER)
      milliseconds -= TIME_BUFFER;

   game->clock.time_left = milliseconds;
   game->move_clock[game->moves_played] = game->clock.time_left;
   set_time_for_game(&game->clock);
}

static void set_timecontrol_from_string(game_t *game, const char *tcstr)
{
   int moves, milliseconds;
   float minutes, seconds, inc;
   /* Set defaults */
   set_time_per_move(&game->clock, 5000);
   game->clock.movestotc = 0;
   game->clock.movestogo = 0;
   game->clock.time_inc = 0;

   if (strstr(tcstr, ":")) {
      sscanf(tcstr, "%d %g:%g %g", &moves, &minutes, &seconds, &inc);
   } else {
      sscanf(tcstr, "%d %g %g", &moves, &minutes, &inc);
      seconds = 0;
   }
   seconds += minutes*60;
   milliseconds = seconds * 1000;

   /* Reserve a bit of time for when we reach the time control, so we don't get penalised by a time
    * loss. We don't need this for a Fischer clock.
    */
   if (inc == 0 && milliseconds > TIME_BUFFER)
      milliseconds -= TIME_BUFFER;

   game->clock.movestotc = moves;
   game->clock.movestogo = moves;
   game->clock.time_inc = inc*1000;
   game->clock.time_left = milliseconds;
   game->move_clock[game->moves_played] = game->clock.time_left;
   set_time_for_game(&game->clock);

   tc_moves = game->clock.movestotc;
   tc_time  = game->clock.time_left;
   tc_inc   = game->clock.time_inc;
}

static void log_xboard_output(const char *msg, ...)
{
   va_list ap;
   static bool newline = true;
   va_start(ap, msg);
   vsnprintf(buf, 65535, msg, ap);
   va_end(ap);

   if (log_f) {
      if (newline)
         fprintf(log_f, "> ");
      newline = false;
      fprintf(log_f, "%s", buf);
      if (strstr(buf, "\n"))
         newline = true;
   }

   printf("%s", buf);
}

static void log_xboard_error(const char *msg, ...)
{
   va_list ap;
   static bool newline = true;
   va_start(ap, msg);
   vsnprintf(buf, 65535, msg, ap);
   va_end(ap);

   if (log_f) {
      if (newline)
         fprintf(log_f, "> ");
      newline = false;
      fprintf(log_f, "%s", buf);
      if (strstr(buf, "\n"))
         newline = true;
   }

   printf("telluser %s", buf);
}

static int max(int x, int y) { return (x>y) ? x : y; }

void initialise(void)
{
   /* Initialise mapping table for board squares to mailbox */
   for (int r=0; r<RANKS; r++) {
      for (int f=0; f<FILES; f++) {
         int sqr  = f+r*FILES;
         int msqr = square_to_mailbox[sqr] = rank_file_to_mailbox(r, f);
         mailbox_to_file[msqr] = f;
         mailbox_to_rank[msqr] = r;
      }
   }

   /* Initialise tropism table */
   memset(tropism, 0, sizeof tropism);
   for (int r1=0; r1<RANKS; r1++) {
      for (int f1=0; f1<FILES; f1++) {
         for (int r2=0; r2<RANKS; r2++) {
            for (int f2=0; f2<FILES; f2++) {
               int sqr1  = f1+r1*FILES;
               int msqr1 = square_to_mailbox[sqr1];
               int sqr2  = f2+r2*FILES;
               int msqr2 = square_to_mailbox[sqr2];
               int index = 256 + msqr1 - msqr2;
               int dist = max(abs(f1-f2), abs(r1-r2));
               tropism[index] = max(tropism[index], dist);
            }
         }
      }
   }

   /* Piece square/centre table */
   int c = RANKS;
   for (int r=0; r<RANKS; r++) {
      for (int f=0; f<FILES; f++) {
         int msqr = square_to_mailbox[f+r*FILES];
         centre_table[msqr] = ((c-1)&~0x1) - (abs(2*r - c + 1)/2 + abs(2*f - c + 1)/2);
         centre_sqr_table[msqr] = -(((r-RANKS/2)*(r-(RANKS-1)/2)+(f-FILES/2)*(f-(FILES-1)/2)));
         advance_table[msqr] = r;
         relative_square[WHITE][msqr] = msqr;
         relative_square[BLACK][msqr] = square_to_mailbox[f+(RANKS-1-r)*FILES];
      }
   }

#if 0
   for (int r=0; r<RANKS; r++) {
      for (int f=0; f<FILES; f++) {
         int msqr = square_to_mailbox[f+r*FILES];
         printf("% 3d ", centre_sqr_table[msqr]);
      }
      printf("\n");
   }
#endif

   /* Square names */
   memset(square_str, 0, sizeof(square_str));
   for (int r=0; r<RANKS; r++) {
      for (int f=0; f<FILES; f++) {
         snprintf(square_str[square_to_mailbox[f+r*FILES]], 4, "%c%d", 'a'+f, 1+r);
      }
   }

   /* Piece types/properties */
   for (int id=0; id<piece_types; id++) {
      piece_info[id].flags = 0;
      for (int n=0; n<MAX_DIRECTIONS; n++) {
         move_description_t md = piece_info[id].move[n];
         if (md.leap ) piece_info[id].flags |= PF_LEAP;
         if (md.slide) piece_info[id].flags |= PF_SLIDE;
      }

      //printf("%2d %10s %02x\n", id, piece_info[id].name, piece_info[id].flags);

      piece_info[id].piece_list_mask = 0ull;
      piece_info[id].piece_list_base = 0;

      /* Pawn */
      if (piece_info[id].wfen == 0 || piece_info[id].bfen == 0) {
         piece_info[id].piece_list_mask = PIECE_LIST_PAWN_MASK;
         piece_info[id].piece_list_base = 0;
         piece_info[id].flags |= PF_PROMOTE;
      } else if (id == 2) {
         piece_info[id].piece_list_mask = PIECE_LIST_ROYAL_MASK;
         piece_info[id].piece_list_base = PIECE_LIST_ROYAL_START;
      } else if (piece_info[id].flags == PF_LEAP) {
         piece_info[id].piece_list_mask = PIECE_LIST_LEAPER_MASK;
         piece_info[id].piece_list_base = PIECE_LIST_LEAPER_START;
      } else if (piece_info[id].flags == PF_SLIDE) {
         piece_info[id].piece_list_mask = PIECE_LIST_SLIDER_MASK;
         piece_info[id].piece_list_base = PIECE_LIST_SLIDER_START;
      } else {
         piece_info[id].piece_list_mask = PIECE_LIST_MIX_MASK;
         piece_info[id].piece_list_base = PIECE_LIST_MIX_START;
      }
   }

   initialise_hash_keys();

   if (piece_types > MAX_PIECE_TYPES) {
      log_xboard_error("*** Too many piece types!\n");
      exit(0);
   }
}

static bool keyboard_input_on_move(game_t *game)
{
   if (deferred[0])  return true;
   if (abort_search) return true;
   static char ponder_input[65536];
   bool input_waiting = keyboard_input_waiting();
   bool read_input    = input_waiting && fgets(ponder_input, sizeof ponder_input, stdin);
   if (read_input) {
      chomp(ponder_input);
      
      if (strstr(ponder_input, "random")) {
         sgenrand(time(NULL));
      } else if (strstr(ponder_input, "easy")) {
      } else if (strstr(ponder_input, "hard")) {
      } else if (strstr(ponder_input, "post")) {
         set_xboard_output_function(game, log_xboard_output);
      } else if (strstr(ponder_input, "nopost")) {
         set_xboard_output_function(game, NULL);
      } else if (strstr(ponder_input, "?") == ponder_input) {
         return true;
      } else if (strstr(ponder_input, "stop") == ponder_input) {
         return true;
      } else if (strstr(ponder_input, "draw")) {
         if (game->draw_count > draw_count)
            log_xboard_output("offer draw\n");
      } else if (strstr(ponder_input, "new")) {
         snprintf(deferred, sizeof deferred, "%s", ponder_input);
         return true;
      } else if (strstr(ponder_input, "otim")) {
      } else if (strstr(ponder_input, "result")) {
      } else if (strstr(ponder_input, "time")) {
      } else if (strstr(ponder_input, "pause")) {
         uint64_t start_pause = get_timer();
         /* Sleep until keyboard input */
         while(!fgets(ponder_input, sizeof ponder_input, stdin) || !strstr(ponder_input, "resume"));
         uint64_t stop_pause = get_timer();
         game->clock.start_time += stop_pause - start_pause;
      } else if (streq(ponder_input, "quit")) {
         if (game)
            end_game(game);
         exit(0);
      } else {
         snprintf(deferred, sizeof deferred, "%s", ponder_input);
         return true;
      }
   }
   return false;
}

static bool keyboard_input_analyse(game_t *game)
{
   if (deferred[0])  return true;
   static char ponder_input[65536];
   bool restart = false;
   if (game->analyse_move != 0) return true;
   if (!game->analysing) return false;

   if (keyboard_input_waiting() && fgets(ponder_input, sizeof ponder_input, stdin)) {
      chomp(ponder_input);
      trim(ponder_input);
      if (strstr(ponder_input, "undo")) {
         if (game->analyse_moves_played) {
            game->analyse_undo++;
            restart = true;
         }
      } else if (strstr(ponder_input, "new")) {
         game->analyse_new = true;
         restart = true;
      } else if (strstr(ponder_input, "exit")) {
         game->analysing = false;
         restart = true;
      } else if (strstr(ponder_input, "setboard")) {
         char *p = strstr(ponder_input, " ");
         while (p && *p && p[0]==' ') p++;
         free((void *)game->analyse_fen);
         game->analyse_fen = strdup(p);
         restart = true;
      } else if (strstr(ponder_input, ".") == ponder_input) {
      } else if (strstr(ponder_input, "hint")) {
      } else if (strstr(ponder_input, "bk")) {
      } else if (streq(ponder_input, "quit")) {
         exit(0);
      } else {
         snprintf(deferred, sizeof deferred, "%s", ponder_input);
         restart = true;
      }
   }
   return restart;
}

int main(int argc, char **argv)
{
   bool xboard_mode = false;
   int time_per_move = 0;
   int nps = -1;
   int depth = MAX_SEARCH_DEPTH;
   bool prompt = stdin_is_terminal();
   bool show_board = true;
   bool san = true;
   bool trapint = true;

   ansi = stdout_is_terminal();

   initialise();

   for (int n = 1; n<argc; n++) {
      if (strstr(argv[n], "-log") || strstr(argv[n], "-newlog")) {
         const char *logfile = "postduif.log";
         if (n+1 < argc && argv[n+1][0] != '-') {
            logfile = argv[n+1];
            n++;
         }
         const char *mode = "a";
         if (strstr(argv[n], "-newlog")) mode = "w";
         if (!log_f) log_f = fopen(logfile, mode);
      } else if (strstr(argv[n], "-") == argv[n]) {
         fprintf(stderr, "Unknown option: %s\n", argv[n]);
         exit(0);
      }
   }

   game_t *game = create_game();
   game->start_fen = strdup(grand_acedrex_fen);
   start_new_game(game);

   /* Calculate attack map of possible attackers for each square */
   int can_attack[32] = {0};
   for (int from = MAILBOX_FIRST; from<255; from++)
   for (int to = MAILBOX_FIRST; to<255; to++) {
      if (game->board.mailbox[from] == B) continue;
      if (game->board.mailbox[to] == B) continue;
      int index = from - to + 256;

      for (int id = 0; id<piece_types; id++) {
         if (piece_info[id].symbol == NULL) continue;

         if (calculate_piece_can_attack(&game->board, id, from, to)) {
            attack_mask[index] |= (1 << id);
            can_attack[id]++;
         }
      }
   }

#if 0
   int correct[32] =  {0};
   int false_positive[32] =  {0};
   int total_false = 0;
   for (int from = MAILBOX_FIRST; from<255; from++)
   for (int to = MAILBOX_FIRST; to<255; to++) {
      if (game->board.mailbox[from] == B) continue;
      if (game->board.mailbox[to] == B) continue;
      int index = from - to + 256;

      for (int id = 0; id<piece_types; id++) {
         if (piece_info[id].symbol == NULL) continue;

         if (!calculate_piece_can_attack(&game->board, id, from, to) && (attack_mask[index] & (1 << id))) {
            false_positive[id]++;
            total_false++;
         } else
            correct[id]++;
      }
   }
   printf("%d false positives\n", total_false);
   for (int id=0; id<piece_types; id++) {
      if (false_positive[id] == 0) continue;
      printf("%10s %d / %d, %d correct\n", piece_info[id].name, false_positive[id], can_attack[id], correct[id]);
   }
#endif




   //for (int n=0; n<board.piece_count[WHITE]; n++) {
      //int id = board.piece_list[n].id;
      //printf("%s%s\n", piece_info[id].symbol, square_str[board.piece_list[n].where]);
   //}

   movelist_t ml = { NULL };

   set_default_output_function(game, log_xboard_output);
   set_xboard_output_function(game, NULL);
   set_error_output_function(game, log_xboard_error);

   /* Turn off buffering for stdout and stdin */
   setvbuf(stdout, NULL, _IONBF, 0);
   setvbuf(stdin, NULL, _IONBF, 0);

   static char input[65536];

   enum {
      PLAY_WHITE = WHITE, PLAY_BLACK = BLACK, FORCE, ANALYSE
   };

   int mode = FORCE;
   while (true) {
      if (show_board) {
         print_board(&game->board);
         printf("%d %d\n", in_check(&game->board, WHITE), in_check(&game->board, BLACK));
      }
      legal_movegen(&ml, &game->board);

      input[0] = 0;
      if (deferred[0]) {
         snprintf(input, sizeof input, "%s", deferred);
         deferred[0] = 0;
      }
      do {
         if (input[0] == 0) {
            if (prompt) printf("#> ");
            if (!fgets(input, sizeof input, stdin))
               exit(0);
         }
         chomp(input);
         trim(input);
         if (input[0] == '#') input[0] = 0;
      } while (input[0] == 0);

      if (strcmp(input, "xboard") == 0 || strcmp(input, "xboard on") == 0) {
         xboard_mode = true;
         show_board = false;
         prompt = false;
         san = false;
         mode = PLAY_BLACK;
         set_xboard_output_function(game, log_xboard_output);
         set_default_output_function(game, NULL);
         printf("\n");
      } else if (strcmp(input, "xboard off") == 0) {
         xboard_mode = false;
         show_board = true;
         prompt = stdin_is_terminal();
         san = true;
         set_xboard_output_function(game, NULL);
         set_default_output_function(game, log_xboard_output);
         printf("\n");
      } else if (strstr(input, "protover") == input) {
         printf("feature variants=\"grandacedrex,grande-acedrex\"\n"
                "feature setboard=1"
                           " time=1"
                         " sigint=0"
                         " colors=0"
                           " ping=1"
                         " memory=1"
                        " analyze=1"
                          " pause=1"
                            " nps=1"
                         " myname=\"%s %s\"\n", "Postduif", "0.14-r"SVNVERSION);
         printf("feature option=\"Draw offer threshold -spin %d 0 1000\"\n", draw_threshold);
         printf("feature option=\"Moves before draw offer (0 to disable) -spin %d 0 1000\"\n", draw_count);
         printf("feature option=\"Resign threshold -spin %d 100 %d\"\n", abs(resign_threshold), CHECKMATE);
         printf("feature option=\"Moves before resigning (0 to disable) -spin %d 0 1000\"\n", resign_count);
         printf("feature option=\"X/WinBoard 4.8 compatible -check %d\"\n", xboard48);
         printf("feature done=1\n");
      } else if (strstr(input, "new") == input) {
         start_new_game(game);
         mode = PLAY_BLACK;
         depth = MAX_SEARCH_DEPTH;
         nps = -1;
      } else if (strstr(input, "setboard") == input) {
         char *fen = strstr(input, " ");
         while (fen && *fen && fen[0]==' ') fen++;

         if (!xboard_mode) mode = FORCE;
         setup_fen(game, fen);
      } else if (strstr(input, "fen") == input) {
         log_xboard_output("%s\n", make_fen_string(game, NULL));
      } else if (strstr(input, "accepted") == input) {
      } else if (strstr(input, "rejected") == input) {
      } else if (strstr(input, "option") == input) {
         if (strstr(input+7, "X/WinBoard 4.8 compatible")) {
            char *s = strstr(input, "=");

            if (s) {
               s++;
               int i = xboard48;
               sscanf(s, "%d", &i);
               xboard48 = i != 0;
            }
            free(game->start_fen);
            if (xboard48)
               game->start_fen = strdup(grand_acedrex_no_castle_fen);
            else
               game->start_fen = strdup(grand_acedrex_fen);

         }
         if (strstr(input+7, "Draw offer threshold")) {
            char *s = strstr(input, "-");
            if (s) {
               int i = draw_threshold;
               sscanf(s+1, "%d", &i);
               draw_threshold = i;
            }
         }
         if (strstr(input+7, "Moves before draw offer (0 to disable)")) {
            char *s = strstr(input, "-");
            if (s) {
               int i = draw_count;
               sscanf(s+1, "%d", &i);
               draw_count = i;
            }
         }
         if (strstr(input+7, "Resign threshold")) {
            char *s = strstr(input, "-");
            if (s) {
               int i = resign_threshold;
               sscanf(s+1, "%d", &i);
               resign_threshold = -abs(i);
            }
         }
         if (strstr(input+7, "Moves before resigning (0 to disable)")) {
            char *s = strstr(input, "-");
            if (s) {
               int i = resign_count;
               sscanf(s+1, "%d", &i);
               resign_count = i;
            }
         }
      } else if (strstr(input, "variant") == input) {
         const char *piece_to_char = "P.CR.............A..U...G..LKp.cr.............a..u...g..lk";
         if (xboard48)
            piece_to_char = "P.CR........G....A.LUKp.cr........g....a.luk";
         printf("setup (%s) 12x12+0_fairy %s\n", piece_to_char, game->start_fen);
         for (int n=0; n<piece_types; n++) {
            if (piece_info[n].wfen == 0) continue;
            if (xboard48 && n == 2) {
               printf("piece K& K\n");
               continue;
            }
            printf("piece %c& %s\n", piece_info[n].wfen, piece_info[n].betza);
         }
         start_new_game(game);
      } else if (strstr(input, "memory") == input) {
         size_t size = game->default_hash_size * sizeof(hash_table_entry_t);
         sscanf(input + 7, "%zd", &size);
         set_transposition_table_size(game, size / sizeof(hash_table_entry_t));
      } else if (strcmp(input, "moves") == 0) {
         print_movelist(&ml, &game->board);
      } else if (strstr(input, "prompt on") == input) {
         prompt = true;
      } else if (strstr(input, "prompt off") == input) {
         prompt = false;
         printf("\n");
      } else if (strstr(input, "board on") == input) {
         show_board = true;
      } else if (strstr(input, "board off") == input) {
         show_board = false;
      } else if (strstr(input, "san on") == input) {
         san = true;
      } else if (strstr(input, "san off") == input) {
         san = false;
      } else if (strstr(input, "ansi on") == input) {
         ansi = true;
      } else if (strstr(input, "ansi off") == input) {
         ansi = false;
      } else if (strcmp(input, "board") == 0) {
         print_board(&game->board);
      } else if (strcmp(input, "eval") == 0) {
         printf("%d\n", evaluate(game));
      } else if (streq(input, "settings")) {
         printf("Xboard mode     : %s\n",   xboard_mode ? "on" : "off");
         printf("\n");
         printf("Prompt          : %s\n",   prompt     ? "on" : "off");
         printf("Board           : %s\n",   show_board ? "on" : "off");
         printf("SAN notation    : %s\n",   san        ? "on" : "off");
         printf("Trap SIGINT     : %s\n",   trapint    ? "on" : "off");
         printf("\n");
         printf("Time control    : %d moves in %ds + %dms\n", tc_moves, tc_time, tc_inc);
         printf("Time per move   : %ds\n", time_per_move);
         printf("Max depth       : %d\n", depth);
         printf("Max nodes       : %llu\n", (unsigned long long) (game ?  game->clock.max_nodes : 0));
      } else if (strcmp(input, "piecelist") == 0) {
         printf("0x%016"PRIx64"\n0x%016"PRIx64"\n", game->board.live_pieces[WHITE], game->board.live_pieces[BLACK]);
      } else if (strcmp(input, "mailbox") == 0) {
         for (int s = 0; s<MAILBOX_SIZE; s++) {
            printf("0x%02x ", game->board.mailbox[s]);
            //int r = s / MAILBOX_FILES;
            //int f = s % MAILBOX_FILES;
            //printf("%3d ", r-f);

            if ((s % MAILBOX_FILES) == MAILBOX_FILES-1) printf("\n");
            if (s == MAILBOX_SIZE-1) printf("\n");
         }
      } else if (strcmp(input, "undo") == 0) {
         if (game->draw_count) game->draw_count--;
         if (game->resign_count) game->resign_count--;
         takeback(game);
         game->clock.time_left = game->move_clock[game->moves_played];
      } else if (strcmp(input, "remove") == 0 || strcmp(input, "takeback") == 0) {
         if (game->draw_count) game->draw_count--;
         if (game->resign_count) game->resign_count--;
         takeback(game);
         takeback(game);
         game->clock.time_left = game->move_clock[game->moves_played];
      } else if (strstr(input, "analyze") || strstr(input, "analyse")) {
         mode = ANALYSE;
         game->analysing = true;
      } else if (strstr(input, "force") == input) {
         game->draw_count = 0;
         game->resign_count = 0;
         mode = FORCE;
      } else if (strstr(input, "perft") == input) {
         if (game) {
            int depth = 6;
            int root = 0;
            char *s = input + 5;
            while (*s && isspace(*s)) s++;
            if (*s) {
               sscanf(s, "%d", &depth);
               while(*s && isdigit(*s)) s++;
            }
            while (*s && isspace(*s)) s++;
            if (*s) {
               sscanf(s, "%d", &root);
            }
#ifdef __unix__
            if (trapint) old_signal_handler = signal(SIGINT, interrupt_computer);
#endif
            abort_search = false;
            uint64_t t = get_timer();
            for (int n = 1; n<depth+1; n++) {
               uint64_t nodes = perft(game, n, root);
               uint64_t tt = get_timer();

               if (tt == t) tt++;

               if (abort_search) break;
               printf("%2d %10lld %5.2f %12.2fnps\n", n, (long long int)nodes,
                     (double)((tt - t)/1000000.0), (double)(nodes*1.0e6/(tt-t)));

               t = tt;
            }
            if (abort_search) printf("*** Aborted\n");
#ifdef __unix__
            if (trapint) signal(SIGINT, old_signal_handler);
#endif
         }
      } else if (strstr(input, "sd ") == input) {
         sscanf(input, "sd %d", &depth);
      } else if (strstr(input, "go")) {
         mode = game->board.side_to_move;
         if (game) {
            /* Correctly set the number of moves to go */
            if (game->clock.movestotc) {
               size_t moves_played = game->moves_played / 2;
               if (mode == PLAY_WHITE)
                  moves_played += (game->moves_played % 2);
               moves_played %= game->clock.movestotc;
               game->clock.movestogo = game->clock.movestotc - (int)moves_played;
            }
         }
      } else if (strstr(input, "st inf")) {
         if (game) set_infinite_time(&game->clock);
      } else if (strstr(input, "st ")) {
         float tpm = 5;
         sscanf(input+3, "%g", &tpm);
         time_per_move = tpm * 1000;
         set_time_per_move(&game->clock, time_per_move);
         if (nps > 0) game->clock.max_nodes = nps * time_per_move;
      } else if (strstr(input, "computer") == input) {
      } else if (strstr(input, "name") == input) {
      } else if (strstr(input, "rating") == input) {
      } else if (strstr(input, "random")) {
         sgenrand((unsigned int)time(NULL));
         game->random_ok = true;
      } else if (strstr(input, "attackers") == input) {
         char *sqr_str = input + 10;
         while (*sqr_str && isspace(*sqr_str)) sqr_str++;

         for (int square = 0; square<MAILBOX_SIZE; square++) {
            if (square_str[square] && strcmp(square_str[square], sqr_str) == 0) {
               printf("Atttackers for '%s'\n", square_str[square]);
               for (side_t side = WHITE; side <= BLACK; side++) {
                  uint64_t attackers = get_attackers(&game->board, square, side);
                  printf("%s: ", (side == WHITE) ? "WHITE" : "BLACK");
                  while (attackers) {
                     int idx = lsb64(attackers) + side * MAX_PIECES_PER_SIDE;
                     attackers &= attackers-1;
                     printf("%s ", square_str[game->board.piece_list[idx].where]);
                  }
                  printf("\n");
               }
               break;
            }
         }
      } else if (strstr(input, "easy") || streq(input, "ponder off")) {
      } else if (strstr(input, "hard") || streq(input, "ponder on")) {
      } else if (strstr(input, "nopost") == input || strstr(input, "post off") == input) {
         set_xboard_output_function(game, NULL);
         set_default_output_function(game, NULL);
      } else if (strstr(input, "post") == input) {
         if (xboard_mode)
            set_xboard_output_function(game, log_xboard_output);
         else
            set_default_output_function(game, log_xboard_output);
      } else if (strstr(input, "?") == input) {
      } else if (strstr(input, "hint") == input) {
      } else if (strstr(input, "otim") == input) {
      } else if (strstr(input, "ping") == input) {
         printf("pong %s\n", input+5);
      } else if (strstr(input, "draw") == input) {
         /* Process draw offer */
         if (game && game->draw_count > draw_count)
            printf("offer draw\n");
      } else if (strstr(input, "result") == input) {
      } else if (strstr(input, "time") == input) {
         set_time_from_string(game, input+5);
      } else if (strstr(input, "nps")) {
         /* This is a convoluted mess... */
         sscanf(input+4, "%d", &nps);
         /* If we have a fixed time per move, then we can set a maximum number of nodes.
          * Otherwise... things get complicated.
          */
         game->clock.max_nodes = 0;
         if (time_per_move > 0 && nps > 0) game->clock.max_nodes = nps * time_per_move;
      } else if (strstr(input, "maxnodes")) {
         unsigned int nodes = 0;
         sscanf(input+9, "%u", &nodes);
         game->clock.max_nodes = nodes;
      } else if (strstr(input, "level") == input) {
         set_timecontrol_from_string(game, input+6);
      } else if (streq(input, "quit") || streq(input, "exit") || streq(input, "bye")) {
         exit(0);
      } else {
         bool valid = false;
         for (int n = 0; n<ml.num_moves; n++) {
            if (strcmp(input, short_move_string(&game->board, ml.move[n], NULL)) == 0) {
               playmove(game, ml.move[n]);
               valid = true;
               break;
            }
            if (strcmp(input, lan_move_string(&game->board, ml.move[n], NULL)) == 0) {
               playmove(game, ml.move[n]);
               valid = true;
               break;
            }
         }

         if (!valid) {
            if (isdigit(input[1])) printf("Illegal move: %s\n", input);
            else                   printf("Error (unknown command): %s\n", input);
         }
      }

      /* Think and find a move */
      if (mode == game->board.side_to_move) {
         move_t move;

         deferred[0] = 0;
         if (stdin_is_terminal())
            game->check_keyboard = keyboard_input_on_move;

         /* Check whether NPS time control is sane. */
         if (nps>-1 && !time_per_move) {
            printf("telluser Warning: specifying NPS without fixed time per move is not implemented\n");
            game->clock.max_nodes = nps;
         }

         game->resign_threshold = resign_threshold;
         game->draw_threshold   = draw_threshold;
         game->pondering = false;

#ifdef __unix__
         if (trapint) old_signal_handler = signal(SIGINT, interrupt_computer);
#endif
         play_state_t state = think(game, depth);
#ifdef __unix__
         if (trapint) signal(SIGINT, old_signal_handler);
#endif

         switch (state) {
            case SEARCH_OK:

               /* Offer draw of resign, as appropriate */
               if (draw_count>0 && game->draw_count > draw_count) {
                  game->draw_count /= 2;   /* Wait a while before offering again */
                  printf("offer draw\n");
               } else if (resign_count>0 && game->resign_count > resign_count)
                  printf("resign\n");

               move = game->move_list[game->moves_played-1];
               takeback(game);

               /* Offer draw or resign, as appropriate */
               if (draw_count>0 && game->draw_count > draw_count) {
                  game->draw_count /= 2;   /* Wait a while before offering again */
                  log_xboard_output("offer draw\n");
               } else if (resign_count>0 && game->resign_count > resign_count)
                  log_xboard_output("resign\n");

               printf("move %s\n", 
                     san ? short_move_string(&game->board, move, NULL)
                         : lan_move_string(&game->board, move, NULL));
               playmove(game, move);

               if (game->clock.movestogo) {
                  game->clock.movestogo--;
                  if (!game->clock.movestogo)
                     game->clock.movestogo = game->clock.movestotc;
               }
               break;

            case SEARCH_GAME_ENDED_REPEAT:
               printf("1/2-1/2 {Draw by repetition rule}\n");
               break;
            case SEARCH_GAME_ENDED_50_MOVE:
               printf("1/2-1/2 {Draw by 50 move rule}\n");
               break;
            case SEARCH_GAME_ENDED_MATE:
               printf("%s\n", game->board.side_to_move ? "1-0 {White mates}" : "0-1 {Black mates}");
               break;
            case SEARCH_GAME_ENDED_STALEMATE:
               printf("1/2-1/2 {Stalemate}\n");
               break;
            case SEARCH_GAME_ENDED_INSUFFICIENT:
               printf("1/2-1/2 {Insufficient mate potential}\n");
               break;
         }
      }

      if (mode == ANALYSE) {
         bool (*old_keyboard_handler)(struct game_t *game) = game->check_keyboard;
         bool (*old_clock_handler)(const struct chess_clock_t *clock) = game->clock.check_clock;

         game->analyse_fen = NULL;
         game->analyse_new = false;
         game->analyse_undo = 0;
         game->analyse_moves_played = game->moves_played;

         game->check_keyboard = keyboard_input_analyse;
         game->clock.check_clock = NULL;
         analyse(game);
         game->check_keyboard = old_keyboard_handler;
         game->clock.check_clock = old_clock_handler;

         if (game->analyse_new)
            start_new_game(game);
         for (int n=0; n<game->analyse_undo; n++)
            takeback(game);
         if (game->analyse_fen) {
            setup_fen(game, game->analyse_fen);
            free((void *)game->analyse_fen);
            game->analyse_fen = NULL;
         }
      }
   }

   end_game(game);
   return 0;
}
