#include <string.h>
#include "board.h"
#include "hashkey.h"

int square_to_mailbox[FILES*RANKS];
int mailbox_to_file[MAILBOX_SIZE];
int mailbox_to_rank[MAILBOX_SIZE];
char promotion_map[MAILBOX_FILES];
char square_str[MAILBOX_SIZE][4];
int relative_square[NUM_SIDES][MAILBOX_SIZE];

void clear_board(board_t *board)
{
   for (int n=0; n<MAILBOX_SIZE; n++) board->mailbox[n] = B;
   memset(board->piece_list, 0, sizeof(board->piece_list));
   board->live_pieces[WHITE] = board->live_pieces[BLACK] = 0;
   board->fifty_counter = 0;

   for (int r=0; r<RANKS; r++) {
      for (int f=0; f<FILES; f++) {
         int msqr = MAILBOX_FIRST + f + r*MAILBOX_FILES;
         board->mailbox[msqr] = 0;
      }
   }
}

int generate_piece_list_index(uint64_t mask, piece_t id)
{
   int n = piece_info[id].piece_list_base;

   /* Decide where to add the piece to the list */
   while (mask & piece_info[id].piece_list_mask & (1ull << n))
      n++;

   return n;
}

int add_piece_to_list(board_t *board, side_t side, piece_t piece, int square)
{
   int base = MAX_PIECES_PER_SIDE*side;
   int n = generate_piece_list_index(board->live_pieces[side], piece);

   board->piece_list[base+n].id     = piece;
   board->piece_list[base+n].where  = square;
   board->piece_list[base+n].virgin = false;
   board->live_pieces[side] |= 1ull<<n;

   return base+n;
}

void setup_promotions(const board_t *board)
{
   for (int f = MAILBOX_FILE0; f<MAILBOX_FILE0+FILES; f++) {
      int sqr = MAILBOX_RANK0 * MAILBOX_FILES + f;
      promotion_map[f] = board->piece_list[board->mailbox[sqr]-1].id;
      if (promotion_map[f] == 2) /* Skip king */
         promotion_map[f] = board->piece_list[board->mailbox[sqr-1]-1].id;
      //printf("%d %d %s\n", f, sqr, piece_info[promotion_map[f]].symbol);
   }
}

bool is_capture_move(const board_t *board, move_t move)
{
   int to = get_move_to(move);

   return board->mailbox[to] != 0;
}

bool is_promotion_move(const board_t *board, move_t move)
{
   int from = get_move_from(move);
   int to   = get_move_to(move);
   int rank = to / MAILBOX_FILES;
   int id   = board->piece_list[board->mailbox[from]-1].id;

   return (rank == MAILBOX_RANK0 || rank == MAILBOX_RANK0+RANKS-1) && (piece_info[id].flags & PF_PROMOTE);
}

bool is_irreversible_move(const board_t *board, move_t move)
{
   return is_capture_move(board, move) || is_promotion_move(board, move) || get_move_piece(board, move) < 2;
}

/* Get the piece ID that corresponds to the given move, which is either to
 * be played or has just been played. This latter functionality exploits
 * the fact that the "from" square is now empty.
 */
int get_moved_piece(const board_t *board, move_t move)
{
   int sqr = get_move_from(move);
   if (board->mailbox[sqr] == 0)
      sqr = get_move_to(move);
   return piece_id_on_square(board, sqr);
}

int get_move_piece(const board_t *board, move_t move)
{
   return piece_id_on_square(board, get_move_from(move));
}

void makemove(board_t *board, move_t move, unmake_info_t *ui)
{
   int from = get_move_from(move);
   int to   = get_move_to(move);

   ui->piece  = board->mailbox[from];
   ui->victim = board->mailbox[to];
   ui->hash   = board->hash;
   ui->virgin = board->piece_list[ui->piece-1].virgin;
   ui->piece_id = board->piece_list[ui->piece-1].id;
   ui->live[WHITE] = board->live_pieces[WHITE];
   ui->live[BLACK] = board->live_pieces[BLACK];
   ui->fifty_counter = board->fifty_counter;
   ui->check = board->check;

   board->hash ^= get_hash_key(board->side_to_move, board->piece_list[ui->piece-1].id, from);

   /* Handle promotions */
   if (is_promotion_move(board, move)) {
      board->piece_list[ui->piece-1].where = 0;
      int side = (ui->piece-1 >= MAX_PIECES_PER_SIDE);
      int idx  = (ui->piece-1) & (MAX_PIECES_PER_SIDE-1);
      board->live_pieces[side] ^= (1ull << idx);

      int id = promotion_map[to % MAILBOX_FILES];
      board->mailbox[to] = add_piece_to_list(board, side, id, to)+1;
      board->hash ^= get_hash_key(board->side_to_move, id, to);
   } else {
      board->mailbox[to] = ui->piece;
      board->piece_list[ui->piece-1].virgin = false;
      board->piece_list[ui->piece-1].where = to;
      board->hash ^= get_hash_key(board->side_to_move, board->piece_list[ui->piece-1].id, to);
   }

   board->mailbox[from] = 0;

   if (ui->victim) {
      ui->victim_id = board->piece_list[ui->victim-1].id;
      board->piece_list[ui->victim-1].where = 0;
      int side = (ui->victim-1 >= MAX_PIECES_PER_SIDE);
      int idx  = (ui->victim-1) & (MAX_PIECES_PER_SIDE-1);
      assert(board->live_pieces[side] & (1ull << idx));
      board->live_pieces[side] ^= (1ull << idx);
      board->hash ^= get_hash_key(side, board->piece_list[ui->victim-1].id, to);
   }

   board->side_to_move = next_side(board->side_to_move);
   board->hash ^= get_stm_hash_key();
}

void unmakemove(board_t *board, move_t move, unmake_info_t *ui)
{
   int from = get_move_from(move);
   int to   = get_move_to(move);

   if (ui->victim) {
      board->piece_list[ui->victim-1].where = to;
      board->piece_list[ui->victim-1].id = ui->victim_id;
   }
   board->piece_list[ui->piece-1].where = from;
   board->piece_list[ui->piece-1].id = ui->piece_id;
   board->fifty_counter = ui->fifty_counter;

   board->mailbox[from] = ui->piece;
   board->mailbox[to]   = ui->victim;
   board->hash          = ui->hash;
   board->piece_list[ui->piece-1].virgin = ui->virgin;
   board->live_pieces[WHITE] = ui->live[WHITE];
   board->live_pieces[BLACK] = ui->live[BLACK];
   board->check = ui->check;

   board->side_to_move = next_side(board->side_to_move);
}

int rank_file_to_mailbox(int r, int f)
{
   return MAILBOX_FIRST + f + r*MAILBOX_FILES;
}

inline side_t get_side_to_move(const board_t *board)
{
   return board->side_to_move;
}

void set_side_to_move(board_t *board, side_t side)
{
   board->side_to_move = side;
}

side_t piece_colour_on_square(const board_t *board, int square)
{
   assert(board->mailbox[square] > 0);
   assert(board->mailbox[square] < 2*MAX_PIECES_PER_SIDE+1);
   int idx = board->mailbox[square]-1;
   return idx >= MAX_PIECES_PER_SIDE;
}

int piece_id_on_square(const board_t *board, int square)
{
   assert(board->mailbox[square] > 0);
   assert(board->mailbox[square] < 2*MAX_PIECES_PER_SIDE+1);
   int idx = board->mailbox[square]-1;
   return board->piece_list[idx].id;
}

